# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
import tempfile
import zipfile

from . import errors
from ._base import FileBase


class Zip(FileBase):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
        )
        if source_tag:
            raise errors.SnapcraftSourceInvalidOptionError("zip", "source-tag")
        elif source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("zip", "source-branch")
        if source_depth:
            raise errors.SnapcraftSourceInvalidOptionError("zip", "source-depth")

    def provision(self, dst, clean_target=True, keep_zip=False, src=None):
        if src:
            zip = src
        else:
            zip = os.path.join(self.source_dir, os.path.basename(self.source))

        if clean_target:
            tmp_zip = tempfile.NamedTemporaryFile().name
            shutil.move(zip, tmp_zip)
            shutil.rmtree(dst)
            os.makedirs(dst)
            shutil.move(tmp_zip, zip)

        # Workaround for: https://bugs.python.org/issue15795
        with zipfile.ZipFile(zip, "r") as f:
            for info in f.infolist():
                extracted_file = f.extract(info.filename, path=dst)

                # Extract the mode from the file. Note that external_attr is
                # a four-byte value, where the high two bytes represent UNIX
                # permissions and file type bits, and the low two bytes contain
                # MS-DOS FAT file attributes. Keep the mode to permissions
                # only-- no sticky bit, uid bit, or gid bit.
                mode = info.external_attr >> 16 & 0x1FF

                # If the zip file was created on a non-unix system, it's
                # possible for the mode to end up being zero. That makes it
                # pretty useless, so ignore it if so.
                if mode:
                    os.chmod(extracted_file, mode)

        if not keep_zip:
            os.remove(zip)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from xdg import BaseDirectory


class SnapcraftCache:
    """Generic cache base class.

    This class is responsible for cache location, notification and pruning.
    """

    def __init__(self):
        self.cache_root = os.path.join(BaseDirectory.xdg_cache_home, "snapcraft")


class SnapcraftProjectCache(SnapcraftCache):
    """Project specific cache"""

    def __init__(self, *, project_name):
        super().__init__()
        self.project_cache_root = os.path.join(
            self.cache_root, "projects", project_name
        )


class SnapcraftStagePackageCache(SnapcraftCache):
    """Cache specific to stage-packages."""

    def __init__(self):
        super().__init__()
        self.stage_package_cache_root = os.path.join(self.cache_root, "stage-packages")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Sequence, Set

from snapcraft.internal import repo


_MSG_EXTEND_STAGE_PACKAGES = (
    "The {part_name!r} part is missing libraries that are not "
    "included in the snap or base. They can be satisfied by adding the following "
    "entries to the existing stage-packages for this "
    "part:\n{stage_packages}"
)
_MSG_ADD_STAGE_PACKAGES = (
    "The {part_name!r} part is missing libraries that are not "
    "included in the snap or base. They can be satisfied by adding the following "
    "entry for this part\nstage-packages:\n{stage_packages}"
)
_MSG_UNHANDLED_DEPENDENCIES = (
    "This part is missing libraries that cannot be satisfied with "
    "any available stage-packages known to snapcraft:\n{unhandled_list}\n"
    "These dependencies can be satisfied via additional parts or content sharing. "
    "Consider validating configured filesets if this dependency was built."
)


def _get_formatted_list(items: Set[str]) -> str:
    return "".join(["- {}\n".format(s) for s in sorted(items)]).strip()


class MissingDependencyResolver:
    """A missing DT_NEEDED resolver."""

    def __init__(self, elf_files: Sequence[str]) -> None:
        self._stage_packages_dependencies = set()  # type: Set[str]
        self._unhandled_dependencies = set()  # type: Set[str]
        self._process(elf_files)

    def _process(self, elf_files: Sequence[str]) -> None:
        for elf_file in elf_files:
            try:
                stage_package = repo.Repo.get_package_for_file(
                    file_path=os.path.join(os.path.sep, elf_file)
                )
                self._stage_packages_dependencies.add(stage_package)
            except repo.errors.FileProviderNotFound:
                self._unhandled_dependencies.add(elf_file)

    def print_resolutions(
        self, *, part_name: str, stage_packages_exist: bool, echoer
    ) -> None:
        if not self._stage_packages_dependencies and not self._unhandled_dependencies:
            return

        if self._stage_packages_dependencies:
            stage_packages_list = _get_formatted_list(self._stage_packages_dependencies)
            if stage_packages_exist:
                echoer.warning(
                    _MSG_EXTEND_STAGE_PACKAGES.format(
                        part_name=part_name, stage_packages=stage_packages_list
                    )
                )
            else:
                echoer.warning(
                    _MSG_ADD_STAGE_PACKAGES.format(
                        part_name=part_name, stage_packages=stage_packages_list
                    )
                )

        if self._unhandled_dependencies:
            unhandled_list = _get_formatted_list(self._unhandled_dependencies)
            echoer.warning(
                _MSG_UNHANDLED_DEPENDENCIES.format(
                    part_name=part_name, unhandled_list=unhandled_list
                )
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import urllib.parse
from time import sleep
from typing import Dict, Iterable, List, TextIO, Union

import pymacaroons
import requests

import snapcraft
from snapcraft import config
from snapcraft.internal.indicators import download_requests_stream

from . import logger
from . import _upload
from . import constants
from . import errors

from ._sso_client import SSOClient
from ._snap_index_client import SnapIndexClient
from ._sca_client import SCAClient
from ._up_down_client import UpDownClient


class StoreClient:
    """High-level client for the V2.0 API SCA resources."""

    def __init__(self) -> None:
        super().__init__()
        self.conf = config.Config()
        self.sso = SSOClient(self.conf)
        self.cpi = SnapIndexClient(self.conf)
        self.updown = UpDownClient(self.conf)
        self.sca = SCAClient(self.conf)

    def login(
        self,
        email: str,
        password: str,
        one_time_password: str = None,
        acls: Iterable[str] = None,
        channels: Iterable[str] = None,
        packages: Iterable[Dict[str, str]] = None,
        expires: str = None,
        config_fd: TextIO = None,
        save: bool = True,
    ) -> None:
        """Log in via the Ubuntu One SSO API."""
        if acls is None:
            acls = [
                "package_access",
                "package_manage",
                "package_push",
                "package_register",
                "package_release",
                "package_update",
            ]

        if config_fd:
            self.conf.load(config_fd=config_fd)
        else:
            # Ask the store for the needed capabilities to be associated with
            # the macaroon.
            macaroon = self.sca.get_macaroon(acls, packages, channels, expires)
            caveat_id = self._extract_caveat_id(macaroon)
            unbound_discharge = self.sso.get_unbound_discharge(
                email, password, one_time_password, caveat_id
            )
            # The macaroon has been discharged, save it in the config
            self.conf.set("macaroon", macaroon)
            self.conf.set("unbound_discharge", unbound_discharge)
            self.conf.set("email", email)

        if save:
            self.conf.save()

    def _extract_caveat_id(self, root_macaroon):
        macaroon = pymacaroons.Macaroon.deserialize(root_macaroon)
        # macaroons are all bytes, never strings
        sso_host = urllib.parse.urlparse(self.sso.root_url).netloc
        for caveat in macaroon.caveats:
            if caveat.location == sso_host:
                return caveat.caveat_id
        else:
            raise errors.InvalidCredentialsError("Invalid root macaroon")

    def logout(self):
        self.conf.clear()
        self.conf.save()

    def _refresh_if_necessary(self, func, *args, **kwargs):
        """Make a request, refreshing macaroons if necessary."""
        try:
            return func(*args, **kwargs)
        except errors.StoreMacaroonNeedsRefreshError:
            unbound_discharge = self.sso.refresh_unbound_discharge(
                self.conf.get("unbound_discharge")
            )
            self.conf.set("unbound_discharge", unbound_discharge)
            self.conf.save()
            return func(*args, **kwargs)

    def whoami(self):
        """Return user relevant login information."""
        account_data = {}

        for k in ("email", "account_id"):
            value = self.conf.get(k)
            if not value:
                account_info = self.get_account_information()
                value = account_info.get(k, "unknown")
                self.conf.set(k, value)
                self.conf.save()
            account_data[k] = value

        return account_data

    def acl(self) -> Dict[str, Union[List[str], str]]:
        """Return permissions for the logged-in user."""

        acl_data = {}

        acl_info = self.verify_acl()
        for key in ("snap_ids", "channels", "permissions", "expires"):
            acl_data[key] = acl_info.get(key)

        return acl_data

    def get_snap_name_for_id(self, snap_id: str) -> str:
        declaration_assertion = self.cpi.get_assertion("snap-declaration", snap_id)
        return declaration_assertion["headers"]["snap-name"]

    def verify_acl(self) -> Dict[str, Union[List[str], str]]:
        return self._refresh_if_necessary(self.sca.verify_acl)

    def get_account_information(self):
        return self._refresh_if_necessary(self.sca.get_account_information)

    def register_key(self, account_key_request):
        return self._refresh_if_necessary(self.sca.register_key, account_key_request)

    def register(self, snap_name: str, is_private: bool = False, store_id: str = None):
        return self._refresh_if_necessary(
            self.sca.register,
            snap_name,
            is_private=is_private,
            store_id=store_id,
            series=constants.DEFAULT_SERIES,
        )

    def push_precheck(self, snap_name):
        return self._refresh_if_necessary(self.sca.snap_push_precheck, snap_name)

    def push_snap_build(self, snap_id, snap_build):
        return self._refresh_if_necessary(self.sca.push_snap_build, snap_id, snap_build)

    def upload(
        self,
        snap_name,
        snap_filename,
        delta_format=None,
        source_hash=None,
        target_hash=None,
        delta_hash=None,
        built_at=None,
    ):
        # FIXME This should be raised by the function that uses the
        # discharge. --elopio -2016-06-20
        if self.conf.get("unbound_discharge") is None:
            raise errors.InvalidCredentialsError(
                "Unbound discharge not in the config file"
            )

        updown_data = _upload.upload_files(snap_filename, self.updown)

        return self._refresh_if_necessary(
            self.sca.snap_push_metadata,
            snap_name,
            updown_data,
            delta_format=delta_format,
            source_hash=source_hash,
            target_hash=target_hash,
            delta_hash=delta_hash,
            built_at=built_at,
        )

    def release(self, snap_name, revision, channels):
        return self._refresh_if_necessary(
            self.sca.snap_release, snap_name, revision, channels
        )

    def get_snap_revisions(self, snap_name, series=None, arch=None):
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        response = self._refresh_if_necessary(
            self.sca.snap_revisions, snap_id, series, arch
        )

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def get_snap_status(self, snap_name, series=None, arch=None):
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        response = self._refresh_if_necessary(
            self.sca.snap_status, snap_id, series, arch
        )

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def close_channels(self, snap_id, channel_names):
        return self._refresh_if_necessary(
            self.sca.close_channels, snap_id, channel_names
        )

    def download(
        self,
        snap_name,
        *,
        risk: str,
        download_path: str,
        track: str = None,
        arch: str = None,
        except_hash=""
    ):
        if arch is None:
            arch = snapcraft.ProjectOptions().deb_arch

        snap_info = self.cpi.get_info(snap_name)
        channel_mapping = snap_info.get_channel_mapping(
            risk=risk, track=track, arch=arch
        )
        if channel_mapping.download.sha3_384 == except_hash:
            return channel_mapping.download.sha3_384

        try:
            channel_mapping.download.verify(download_path)
        except errors.StoreDownloadError:
            self._download_snap(channel_mapping.download, download_path)

        channel_mapping.download.verify(download_path)
        return channel_mapping.download.sha3_384

    def _download_snap(self, download_details, download_path):
        # we only resume when redirected to our CDN since we use internap's
        # special sauce.
        total_read = 0
        probe_url = requests.head(download_details.url)
        if probe_url.is_redirect and "internap" in probe_url.headers["Location"]:
            download_url = probe_url.headers["Location"]
            resume_possible = True
        else:
            download_url = download_details.url
            resume_possible = False

        # HttpAdapter cannot help here as this is a stream.
        # LP: #1617765
        not_downloaded = True
        retry_count = 5
        while not_downloaded and retry_count:
            headers = {}
            if resume_possible and os.path.exists(download_path):
                total_read = os.path.getsize(download_path)
                headers["Range"] = "bytes={}-".format(total_read)
            request = self.cpi.get(download_url, headers=headers, stream=True)
            request.raise_for_status()
            redirections = [h.headers["Location"] for h in request.history]
            if redirections:
                logger.debug(
                    "Redirections for {!r}: {}".format(
                        download_url, ", ".join(redirections)
                    )
                )
            try:
                download_requests_stream(request, download_path, total_read=total_read)
                not_downloaded = False
            except requests.exceptions.ChunkedEncodingError as e:
                logger.debug(
                    "Error while downloading: {!r}. "
                    "Retries left to download: {!r}.".format(e, retry_count)
                )
                retry_count -= 1
                if not retry_count:
                    raise e
                sleep(1)

    def push_assertion(self, snap_id, assertion, endpoint, force=False):
        return self.sca.push_assertion(snap_id, assertion, endpoint, force)

    def get_assertion(self, snap_id, endpoint):
        return self.sca.get_assertion(snap_id, endpoint)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        return self.sca.sign_developer_agreement(latest_tos_accepted)

    def push_metadata(self, snap_name, metadata, force):
        """Push the metadata to the server."""
        account_info = self.get_account_information()
        series = constants.DEFAULT_SERIES
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self._refresh_if_necessary(
            self.sca.push_metadata, snap_id, snap_name, metadata, force
        )

    def push_binary_metadata(self, snap_name, metadata, force):
        """Push the binary metadata to the server."""
        account_info = self.get_account_information()
        series = constants.DEFAULT_SERIES
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self._refresh_if_necessary(
            self.sca.push_binary_metadata, snap_id, snap_name, metadata, force
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import sys
import urllib.parse
import warnings
from typing import Optional, Sequence

from .._base_provider import Provider
from .._base_provider import errors
from ._images import get_image_source
from snapcraft.internal import repo
from snapcraft.internal.errors import SnapcraftEnvironmentError

# LXD is only supported on Linux and causes issues when imported on Windows.
# We conditionally import it and rely on ensure_provider() to check OS before
# using pylxd.
if sys.platform == "linux":
    import pylxd

logger = logging.getLogger(__name__)
# Filter out attribute setting warnings for properties that exist in LXD operations
# but are unhandled in pylxd.
warnings.filterwarnings("ignore", module="pylxd.models.operation")


class LXD(Provider):
    """A LXD provider for snapcraft to execute its lifecycle."""

    _PROJECT_DEVICE_NAME = "snapcraft-project"
    _PROJECT_EXPORTED_PRIME_NAME = "snapcraft-project-prime"

    # Given that we are running snapcraft from the snapcraft snap, which has
    # classic confinement and require using the lxd snap, the lxd and lxc
    # binaries should be found in /snap/bin
    _LXD_BIN = os.path.join(os.path.sep, "snap", "bin", "lxd")
    _LXC_BIN = os.path.join(os.path.sep, "snap", "bin", "lxc")

    @classmethod
    def ensure_provider(cls):
        error_message = None  # type: Optional[str]
        prompt_installable = False

        if sys.platform != "linux":
            error_message = "LXD is not supported on this platform"
        else:
            try:
                if not repo.snaps.SnapPackage.is_snap_installed("lxd"):
                    error_message = (
                        "The LXD snap is required to continue: snap install lxd"
                    )
                    prompt_installable = True
            except repo.errors.SnapdConnectionError:
                error_message = (
                    "snap support is required to continue: "
                    "https://docs.snapcraft.io/installing-snapd/6735"
                )

        if error_message is not None:
            raise errors.ProviderNotFound(
                provider=cls._get_provider_name(),
                prompt_installable=prompt_installable,
                error_message=error_message,
            )

        # If we reach this point, it means the lxd snap is properly setup.
        # Now is the time for additional sanity checks to ensure the provider
        # will work.
        try:
            # TODO: add support for more distributions. Maybe refactor a bit so that Repo behaves
            # similar to a build provider.
            if repo.Repo.is_package_installed("lxd") or repo.Repo.is_package_installed(
                "lxd-client"
            ):
                raise SnapcraftEnvironmentError(
                    (
                        "The {!r} provider does not support having the 'lxd' or "
                        "'lxd-client' deb packages installed. To completely migrate "
                        "to the LXD snap run 'lxd.migrate' and try again."
                    ).format(cls._get_provider_name())
                )
        except repo.errors.NoNativeBackendError:
            pass

    @classmethod
    def setup_provider(cls, *, echoer) -> None:
        repo.snaps.install_snaps(["lxd/latest/stable"])

        try:
            subprocess.check_output([cls._LXD_BIN, "waitready", "--timeout=30"])
        except subprocess.CalledProcessError as call_error:
            raise SnapcraftEnvironmentError(
                "Timeout reached waiting for LXD to start."
            ) from call_error

        try:
            subprocess.check_output([cls._LXD_BIN, "init", "--auto"])
        except subprocess.CalledProcessError as call_error:
            raise SnapcraftEnvironmentError(
                "Failed to initialize LXD. "
                "Try manually initializing before trying again: lxd init --auto."
            ) from call_error

    @classmethod
    def _get_provider_name(cls):
        return "LXD"

    @classmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        return True

    def _run(
        self, command: Sequence[str], hide_output: bool = False
    ) -> Optional[bytes]:
        self._ensure_container_running()

        logger.debug("Running {}".format(" ".join(command)))
        # TODO: use pylxd
        cmd = [self._LXC_BIN, "exec", self.instance_name, "--"] + list(command)
        try:
            if hide_output:
                output = subprocess.check_output(cmd)
            else:
                # output here will be None, and that is OK.
                output = subprocess.check_call(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderExecError(
                provider_name=self._get_provider_name(),
                command=command,
                exit_code=process_error.returncode,
            ) from process_error

        return output

    def _launch(self) -> None:
        config = {
            "name": self.instance_name,
            "source": get_image_source(base=self.project.info.get_build_base()),
        }

        try:
            container = self._lxd_client.containers.create(config, wait=True)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderLaunchError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            ) from lxd_api_error
        container.config["user.user-data"] = self._get_cloud_user_data_string()
        # This is setup by cloud init, but set it here to be on the safer side.
        container.config["environment.SNAPCRAFT_BUILD_ENVIRONMENT"] = "managed-host"
        container.save(wait=True)
        self._container = container

        self._start()

    def _start(self):
        if not self._lxd_client.containers.exists(self.instance_name):
            raise errors.ProviderInstanceNotFoundError(instance_name=self.instance_name)

        if self._container is None:
            self._container = self._lxd_client.containers.get(self.instance_name)

        self._container.sync()
        self._container.config["environment.SNAPCRAFT_HAS_TTY"] = str(
            sys.stdout.isatty()
        )
        # map to the owner of the directory we are eventually going to write the
        # snap to.
        self._container.config["raw.idmap"] = "both {!s} 0".format(
            os.stat(self.project._project_dir).st_uid
        )
        self._container.save(wait=True)

        if self._container.status.lower() != "running":
            try:
                self._container.start(wait=True)
            except pylxd.exceptions.LXDAPIException as lxd_api_error:
                print(self._container.status)
                raise errors.ProviderStartError(
                    provider_name=self._get_provider_name(), error_message=lxd_api_error
                ) from lxd_api_error

        # Ensure cloud init is done
        self.echoer.wrapped("Waiting for cloud-init")
        self._run(command=["cloud-init", "status", "--wait"])

    def _stop(self):
        # If _container is still None here it means creation/starting was not
        # successful.
        if self._container is None:
            return

        self._container.sync()

        if self._container.status.lower() != "stopped":
            try:
                self._container.stop(wait=True)
            except pylxd.exceptions.LXDAPIException as lxd_api_error:
                raise errors.ProviderStopError(
                    provider_name=self._get_provider_name(), error_message=lxd_api_error
                ) from lxd_api_error

    def _push_file(self, *, source: str, destination: str) -> None:
        self._ensure_container_running()

        # TODO: better handling of larger files.
        with open(source, "rb") as source_data:
            source_contents = source_data.read()

        try:
            self._container.files.put(destination, source_contents)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            )

    def __init__(self, *, project, echoer, is_ephemeral: bool = False) -> None:
        super().__init__(project=project, echoer=echoer, is_ephemeral=is_ephemeral)
        self.echoer.warning(
            "The LXD provider is offered as a technology preview for early adopters.\n"
            "The command line interface, container names or lifecycle handling may "
            "change in upcoming releases."
        )
        # This endpoint is hardcoded everywhere lxc/lxd-pkg-snap#33
        lxd_socket_path = "/var/snap/lxd/common/lxd/unix.socket"
        endpoint = "http+unix://{}".format(urllib.parse.quote(lxd_socket_path, safe=""))
        try:
            self._lxd_client = pylxd.Client(endpoint=endpoint)
        except pylxd.client.exceptions.ClientConnectionFailed:
            raise errors.ProviderCommunicationError(
                provider_name=self._get_provider_name(),
                message="cannot connect to the LXD socket ({!r}).".format(
                    lxd_socket_path
                ),
            )
        self._container = None  # type: Optional[pylxd.models.container.Container]

    def create(self) -> None:
        """Create the LXD instance and setup the build environment."""
        self.echoer.info("Launching a container.")
        self.launch_instance()

    def destroy(self) -> None:
        """Destroy the instance, trying to stop it first."""
        self._stop()

    def mount_project(self) -> None:
        if self._PROJECT_DEVICE_NAME in self._container.devices:
            return

        # Resolve the home directory
        home_dir = (
            self._run(command=["printenv", "HOME"], hide_output=True).decode().strip()
        )

        self._container.sync()

        self._container.devices[self._PROJECT_DEVICE_NAME] = {
            "type": "disk",
            "source": self.project._project_dir,
            "path": os.path.join(home_dir, "project"),
        }

        try:
            self._container.save(wait=True)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderMountError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            ) from lxd_api_error

    def _mount_prime_directory(self) -> bool:
        if self._PROJECT_EXPORTED_PRIME_NAME in self._container.devices:
            return True

        # Resolve the home directory
        home_dir = (
            self._run(command=["printenv", "HOME"], hide_output=True).decode().strip()
        )

        self._container.sync()

        self._container.devices[self._PROJECT_EXPORTED_PRIME_NAME] = {
            "type": "disk",
            "source": os.path.join(self.project.prime_dir),
            "path": os.path.join(home_dir, "prime"),
        }

        try:
            self._container.save(wait=True)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderMountError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            ) from lxd_api_error

        return False

    def clean_project(self) -> bool:
        was_cleaned = super().clean_project()
        if was_cleaned:
            if self._container is None:
                self._container = self._lxd_client.containers.get(self.instance_name)
            self._stop()
            self._container.delete(wait=True)
        return was_cleaned

    def pull_file(self, name: str, destination: str, delete: bool = False) -> None:
        self._ensure_container_running()

        # TODO: better handling of larger files.
        try:
            source_data = self._container.files.get(name)
        except pylxd.exceptions.LXDAPIException as lxd_api_error:
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(), error_message=lxd_api_error
            )
        else:
            with open(destination, "wb") as destination_data:
                destination_data.write(source_data)
        if delete and self._container.files.delete_available:
            self._container.files.delete(name)
        if delete and not self._container.files.delete_available:
            logger.warning("File deletion not supported by this LXD version.")

    def shell(self) -> None:
        self._run(command=["/bin/bash"])

    def _ensure_container_running(self) -> None:
        self._container.sync()
        if self._container.status.lower() != "running":
            raise errors.ProviderFileCopyError(
                provider_name=self._get_provider_name(),
                error_message=(
                    "Container is not running, the current state is: {!r]. "
                    "Ensure it has not been modified by external factors and try again"
                ).format(self._container.status),
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shlex
from typing import Any, Dict, Optional
from typing import Sequence  # noqa: F401

from snapcraft.internal.errors import SnapcraftError as _SnapcraftError


class ProviderBaseError(_SnapcraftError):
    """Base Exception for all provider related exceptions."""


class ProviderNotSupportedError(ProviderBaseError):

    fmt = (
        "The {provider!r} provider is not supported, please choose a "
        "different one and try again."
    )

    def __init__(self, *, provider: str) -> None:
        super().__init__(provider=provider)


class ProviderNotFound(ProviderBaseError):

    fmt = "You need {provider!r} set-up to build snaps: {error_message}."

    def __init__(
        self, *, provider: str, prompt_installable: bool, error_message: str
    ) -> None:
        super().__init__(
            provider=provider,
            prompt_installable=prompt_installable,
            error_message=error_message,
        )
        self.prompt_installable = prompt_installable
        self.provider = provider


class _GenericProviderError(ProviderBaseError):

    _FMT_ERROR_MESSAGE_AND_EXIT_CODE = (
        "An error occurred with the instance when trying to {action} with "
        "{provider_name!r}: returned exit code {exit_code!r}: {error_message}.\n"
        "Ensure that {provider_name!r} is setup correctly and try again."
    )

    _FMT_ERROR_MESSAGE = (
        "An error occurred with the instance when trying to {action} with "
        "{provider_name!r}: {error_message}.\n"
        "Ensure that {provider_name!r} is setup correctly and try again."
    )

    _FMT_EXIT_CODE = (
        "An error occurred with the instance when trying to {action} with "
        "{provider_name!r}: returned exit code {exit_code!r}.\n"
        "Ensure that {provider_name!r} is setup correctly and try again."
    )

    def __init__(
        self,
        *,
        provider_name: str,
        action: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        if exit_code is not None and error_message is not None:
            fmt = self._FMT_ERROR_MESSAGE_AND_EXIT_CODE
        elif error_message:
            fmt = self._FMT_ERROR_MESSAGE
        elif exit_code:
            fmt = self._FMT_EXIT_CODE
        else:
            raise RuntimeError("error_message nor exit_code are set")

        self.fmt = fmt

        super().__init__(
            provider_name=provider_name,
            action=action,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderCommunicationError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to communicate with the "
        "{provider_name!r} provider: {message}."
    )

    def __init__(self, *, provider_name: str, message: str) -> None:
        super().__init__(provider_name=provider_name, message=message)


class ProviderLaunchError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="launch",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderStartError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="start",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderStopError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="stop",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderDeleteError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="delete",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderExecError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to execute {command_string!r} with "
        "{provider_name!r}: returned exit code {exit_code!r}."
    )

    def __init__(
        self, *, provider_name: str, command: Sequence[str], exit_code: int
    ) -> None:
        command_string = " ".join(shlex.quote(i) for i in command)
        super().__init__(
            provider_name=provider_name,
            command=command,
            command_string=command_string,
            exit_code=exit_code,
        )


class ProviderShellError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="shell",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderMountError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="mount",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderUnMountError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="unmount",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderFileCopyError(_GenericProviderError):
    def __init__(
        self,
        *,
        provider_name: str,
        error_message: Optional[str] = None,
        exit_code: Optional[int] = None
    ) -> None:
        super().__init__(
            action="copy files",
            provider_name=provider_name,
            error_message=error_message,
            exit_code=exit_code,
        )


class ProviderInfoError(ProviderBaseError):

    fmt = (
        "An error occurred when using {provider_name!r} to "
        "query the status of the instance: returned exit code {exit_code!r}: {stderr!s}."
    )

    def __init__(self, *, provider_name: str, exit_code: int, stderr: bytes) -> None:
        super().__init__(
            provider_name=provider_name, exit_code=exit_code, stderr=stderr.decode()
        )


class ProviderInstanceNotFoundError(ProviderBaseError):

    fmt = "Cannot find an instance named {instance_name!r}."

    def __init__(self, *, instance_name: str) -> None:
        super().__init__(instance_name=instance_name)


class ProviderInfoDataKeyError(ProviderBaseError):

    fmt = (
        "The data returned by {provider_name!r} was not expected. "
        "It is missing a required key {missing_key!r} in {data!r}."
    )

    def __init__(
        self, *, provider_name: str, missing_key: str, data: Dict[str, Any]
    ) -> None:
        super().__init__(
            provider_name=provider_name, missing_key=missing_key, data=data
        )


class ProviderBadDataError(ProviderBaseError):

    fmt = (
        "The data returned by {provider_name!r} was not expected "
        "or in the wrong format: {data!r}."
    )

    def __init__(self, *, provider_name: str, data: str) -> None:
        super().__init__(provider_name=provider_name, data=data)


class ProviderMultipassDownloadFailed(ProviderBaseError):
    fmt = (
        "Failed to download Multipass: {message!r}\n"
        "Please install manually. You can find the latest release at:\n"
        "https://multipass.run"
    )

    def __init__(self, message):
        super().__init__(message=message)


class ProviderMultipassInstallationFailed(ProviderBaseError):
    fmt = (
        "Failed to install Multipass: {message!r}\n"
        "Please install manually. You can find the latest release at:\n"
        "https://multipass.run"
    )

    def __init__(self, message):
        super().__init__(message=message)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional, Set

from . import channels
from . import constants
from . import errors


# FIXME Obtain link to docs
"""
This module holds representations for results from the status API provided by
the Snap Store.
"""


class SnapStatusChannelDetails:
    def __repr__(self) -> str:
        return "<SnapStatusChannelDetails: {} on {}>".format(
            self._snap_name, self._arch
        )

    def __iter__(self):
        for attr in ["arch", "revision", "version"]:
            yield self.__getattribute__(attr)

    def __init__(self, *, snap_name: str, arch: str, payload: Dict[str, Any]) -> None:
        self._snap_name = snap_name
        self._arch = arch
        self._payload = payload

    @property
    def arch(self) -> str:
        return self._arch

    @property
    def channel(self) -> str:
        return self._payload["channel"]

    @property
    def info(self) -> str:
        return self._payload["info"]

    @property
    def version(self) -> Optional[str]:
        return self._payload.get("version")

    @property
    def revision(self) -> Optional[int]:
        return self._payload.get("revision")


class SnapStatusTrackDetails:
    def __repr__(self) -> str:
        return "<SnapStatusTrackDetails: {} on {}>".format(self._snap_name, self.track)

    def __init__(self, *, snap_name: str, track: str, payload: Dict[str, Any]) -> None:
        self._snap_name = snap_name
        self._track = track
        self._payload = payload
        self._arch_channels = dict()  # type: Dict[str, List[SnapStatusChannelDetails]]

    @property
    def track(self) -> str:
        return self._track

    def get_arches(self) -> List[str]:
        return [k for k in self._payload.keys()]

    def get_channel(
        self, *, risk: str, arch: str, branch: Optional[str] = None
    ) -> SnapStatusChannelDetails:
        if arch not in self._arch_channels:
            try:
                self._arch_channels[arch] = [
                    SnapStatusChannelDetails(
                        snap_name=self._snap_name, arch=arch, payload=c
                    )
                    for c in self._payload[arch]
                ]
            except KeyError:
                raise errors.ChannelNotAvailableOnArchError(
                    snap_name=self._snap_name,
                    channel=channels.Channel.from_channel_tuple(
                        track=self.track, risk=risk, branch=branch
                    ),
                    arch=arch,
                )

        if branch:
            channel = "{}/{}".format(risk, branch)
        else:
            channel = risk

        c = [c for c in self._arch_channels[arch] if channel == c.channel]
        # We assume the API will not return duplicate mappings
        try:
            return c[0]
        except IndexError:
            raise errors.ChannelNotAvailableOnArchError(
                snap_name=self._snap_name,
                channel=channels.Channel.from_channel_tuple(
                    track=self.track, risk=risk, branch=branch
                ),
                arch=arch,
            )


class SnapStatus:
    def __repr__(self) -> str:
        return "<SnapStatus: {}>".format(self._snap_name)

    def __init__(self, *, snap_name: str, payload: Dict[str, Any]) -> None:
        self._snap_name = snap_name
        self._payload = payload

    def get_tracks(self) -> List[str]:
        return [k for k in self._payload["channel_map_tree"].keys()]

    def get_track(
        self, track: str, *, device_series=constants.DEFAULT_SERIES
    ) -> SnapStatusTrackDetails:
        return SnapStatusTrackDetails(
            snap_name=self._snap_name,
            track=track,
            payload=self._payload["channel_map_tree"][track][device_series],
        )

    def get_channel_set(
        self, channel: channels.Channel
    ) -> Set[SnapStatusChannelDetails]:
        track = self.get_track(channel.track)
        channel_selection = [
            track.get_channel(risk=channel.risk, branch=channel.branch, arch=arch)
            for arch in track.get_arches()
        ]

        # Make sure this is a valid build set (has released revisions for every channel/arch)
        channel_outliers = [c for c in channel_selection if c.revision is None]
        if channel_outliers:
            raise errors.InvalidChannelSet(
                snap_name=self._snap_name,
                channel=channel,
                channel_outliers=channel_outliers,
            )

        return set(channel_selection)



import itertools
from time import sleep
from threading import Thread
from queue import Queue

from progressbar import AnimatedMarker, ProgressBar, UnknownLength

import requests

from . import constants
from . import errors


class StatusTracker:

    __messages = {
        "being_processed": "Processing...",
        "ready_to_release": "Ready to release!",
        "need_manual_review": "Will need manual review...",
        "processing_upload_delta_error": "Error while processing delta...",
        "processing_error": "Error while processing...",
    }

    __error_codes = {
        "processing_error",
        "processing_upload_delta_error",
        "need_manual_review",
    }

    def __init__(self, status_details_url):
        self.__status_details_url = status_details_url

    def track(self):
        queue = Queue()
        thread = Thread(target=self._update_status, args=(queue,))
        thread.start()

        widgets = ["Processing...", AnimatedMarker()]
        progress_indicator = ProgressBar(widgets=widgets, maxval=UnknownLength)
        progress_indicator.start()

        content = {}
        for indicator_count in itertools.count():
            progress_indicator.update(indicator_count)
            if not queue.empty():
                content = queue.get()
                if isinstance(content, Exception):
                    raise content
            if content.get("processed"):
                break
            else:
                widgets[0] = self._get_message(content)
            sleep(0.1)
        progress_indicator.finish()
        # Print at the end to avoid a left over spinner artifact
        print(self._get_message(content))

        self.__content = content

        return content

    def raise_for_code(self):
        if self.__content["code"] in self.__error_codes:
            raise errors.StoreReviewError(self.__content)

    def _get_message(self, content):
        try:
            return self.__messages.get(content["code"], content["code"])
        except KeyError:
            return self.__messages.get("being_processed")

    def _update_status(self, queue):
        for content in self._get_status():
            queue.put(content)
            if content["processed"]:
                break
            sleep(constants.SCAN_STATUS_POLL_DELAY)

    def _get_status(self):
        connection_errors_allowed = 10
        while True:
            try:
                content = requests.get(self.__status_details_url).json()
            except (requests.ConnectionError, requests.HTTPError) as e:
                if not connection_errors_allowed:
                    yield e
                content = {"processed": False, "code": "being_processed"}
                connection_errors_allowed -= 1
            yield content



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
import copy
import jsonschema
import importlib
import logging
import os
import pkgutil
from typing import Any, Dict, List, Set, Type  # noqa: F401

from .. import errors
from ._extension import Extension
from snapcraft.project import errors as project_errors

logger = logging.getLogger(__name__)

# Extensions are evaluated before the snapcraft.yaml is validated, so we need to handle
# our own subset of validation ourselves (covering only the usage of extensions).
extension_schema = {
    "type": "array",
    "minitems": 1,
    "uniqueItems": True,
    "items": {"type": "string"},
}


def apply_extensions(yaml_data: Dict[str, Any]) -> Dict[str, Any]:
    """Apply all extensions.

    :param dict yaml_data: Loaded, unprocessed snapcraft.yaml
    :returns: Modified snapcraft.yaml data with extensions applied
    """
    # Don't modify the dict passed in
    yaml_data = copy.deepcopy(yaml_data)
    original_yaml_data = copy.deepcopy(yaml_data)
    base = yaml_data.get("base")

    # Mapping of extension names to set of app names to which the extension needs to be
    # applied.
    declared_extensions = collections.defaultdict(set)  # type: Dict[str, Set[str]]

    for app_name, app_definition in yaml_data.get("apps", dict()).items():
        extension_names = app_definition.get("extensions", [])
        _validate_extension_format(extension_names)

        for extension_name in extension_names:
            declared_extensions[extension_name].add(app_name)

        # Now that we've saved the app -> extension relationship, remove the property
        # from this app's declaration in the YAML.
        with contextlib.suppress(KeyError):
            del yaml_data["apps"][app_name]["extensions"]

    # Process extensions in a consistent order
    for extension_name in sorted(declared_extensions.keys()):
        extension = _load_extension(base, extension_name, original_yaml_data)
        _apply_extension(
            yaml_data, declared_extensions[extension_name], extension_name, extension
        )

    return yaml_data


def find_extension(extension_name: str) -> Type[Extension]:
    """Find and return the extension class with the given name.

    :param str extension_name: The name of the extension to load

    :returns: Extension subclass responsible for extension.
    :rtype: type(Extension)
    :raises: errors.ExtensionNotFoundError if the extension is not found
    """
    if extension_name.startswith("_"):
        raise errors.ExtensionNotFoundError(extension_name)

    try:
        extension_module = importlib.import_module(
            "snapcraft.internal.project_loader._extensions.{}".format(
                extension_name.replace("-", "_")
            )
        )
    except ImportError:
        raise errors.ExtensionNotFoundError(extension_name)

    # The extension module requires a class named ExtensionImpl.
    return getattr(extension_module, "ExtensionImpl")


def supported_extension_names() -> List[str]:
    """Return a list of supported extension names.

    :returns: List of supported extension names.
    :rtype: list
    """

    extension_names = []  # type: List[str]
    for _, modname, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
        # Only add non-private modules/packages to the extension list
        if not modname.startswith("_"):
            extension_names.append(modname)

    return extension_names


def _load_extension(
    base: str, extension_name: str, yaml_data: Dict[str, Any]
) -> Extension:
    extension_class = find_extension(extension_name)

    # Hand the extension a copy of the yaml data so the only way they can modify it is
    # by going through the extension API.
    return extension_class(
        extension_name=extension_name, yaml_data=copy.deepcopy(yaml_data)
    )


def _apply_extension(
    yaml_data: Dict[str, Any],
    app_names: Set[str],
    extension_name: str,
    extension: Extension,
):
    # Apply the root components of the extension (if any)
    root_extension = extension.root_snippet
    for property_name, property_value in root_extension.items():
        yaml_data[property_name] = _apply_extension_property(
            yaml_data.get(property_name), property_value
        )

    # Apply the app-specific components of the extension (if any)
    app_extension = extension.app_snippet
    for app_name in app_names:
        app_definition = yaml_data["apps"][app_name]
        for property_name, property_value in app_extension.items():
            app_definition[property_name] = _apply_extension_property(
                app_definition.get(property_name), property_value
            )

    # Next, apply the part-specific components
    part_extension = extension.part_snippet
    parts = yaml_data["parts"]
    for part_name, part_definition in parts.items():
        for property_name, property_value in part_extension.items():
            part_definition[property_name] = _apply_extension_property(
                part_definition.get(property_name), property_value
            )

    # Finally, add any parts specified in the extension
    for part_name, part_definition in extension.parts.items():
        # If a extension part name clashes with a part that already exists, error.
        if part_name in parts:
            raise errors.ExtensionPartConflictError(extension_name, part_name)

        parts[part_name] = part_definition


def _apply_extension_property(existing_property: Any, extension_property: Any):
    if existing_property:
        # If the property is not scalar, merge them
        if isinstance(existing_property, list) and isinstance(extension_property, list):
            return _merge_lists(existing_property, extension_property)
        elif isinstance(existing_property, dict) and isinstance(
            extension_property, dict
        ):
            for key, value in extension_property.items():
                existing_property[key] = _apply_extension_property(
                    existing_property.get(key), value
                )
            return existing_property
        return existing_property

    return extension_property


def _merge_lists(list1: List[str], list2: List[str]) -> List[str]:
    """Merge two lists while maintaining order and removing duplicates."""
    seen = set()  # type: Set[str]
    merged = list()  # type: List[str]

    for item in list1 + list2:
        if item not in seen:
            seen.add(item)
            merged.append(item)

    return merged


def _validate_extension_format(extension_names):
    if extension_names is not None:
        format_check = jsonschema.FormatChecker()
        try:
            jsonschema.validate(
                extension_names, extension_schema, format_checker=format_check
            )
        except jsonschema.ValidationError as e:
            raise project_errors.YamlValidationError(
                "The 'extensions' property does not match the required schema: {}".format(
                    project_errors.YamlValidationError.from_validation_error(e).message
                )
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from . import errors
from ._base import Base


class Bazaar(Base):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
        silent=False,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
            "bzr",
        )
        if source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("bzr", "source-branch")
        if source_depth:
            raise errors.SnapcraftSourceInvalidOptionError("bzr", "source-depth")
        if source_tag and source_commit:
            raise errors.SnapcraftSourceIncompatibleOptionsError(
                "bzr", ["source-tag", "source-commit"]
            )
        if source_checksum:
            raise errors.SnapcraftSourceInvalidOptionError("bzr", "source-checksum")

        self._call_kwargs = {}
        if silent:
            self._call_kwargs["stdout"] = subprocess.DEVNULL
            self._call_kwargs["stderr"] = subprocess.DEVNULL

    def pull(self):
        tag_opts = []
        if self.source_tag:
            tag_opts = ["-r", "tag:" + self.source_tag]
        if self.source_commit:
            tag_opts = ["-r", self.source_commit]
        if os.path.exists(os.path.join(self.source_dir, ".bzr")):
            cmd = (
                [self.command, "pull"] + tag_opts + [self.source, "-d", self.source_dir]
            )
        else:
            os.rmdir(self.source_dir)
            cmd = [self.command, "branch"] + tag_opts + [self.source, self.source_dir]

        self._run(cmd, **self._call_kwargs)
        self.source_details = self._get_source_details()

    def _get_source_details(self):
        tag = self.source_tag
        commit = self.source_commit

        if not tag:
            if os.path.exists(self.source_dir):
                commit = self._run_output(["bzr", "revno", self.source_dir])

        branch = None
        source = self.source

        return {
            "source-commit": commit,
            "source-branch": branch,
            "source": source,
            "source-tag": tag,
        }



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from snapcraft.internal import errors


class Step:
    def __init__(self, name: str, clean_if_dirty: bool) -> None:
        self.name = name
        self.clean_if_dirty = clean_if_dirty
        self.__order = None  # type: int

    @property
    def _order(self) -> int:
        if self.__order is None:
            try:
                self.__order = STEPS.index(self)
            except ValueError:
                raise errors.InvalidStepError(self.name)
        return self.__order

    def previous_step(self) -> "Step":
        if self._order > 0:
            return STEPS[self._order - 1]
        else:
            return None

    def next_step(self) -> "Step":
        try:
            return STEPS[self._order + 1]
        except IndexError:
            return None

    def previous_steps(self) -> List["Step"]:
        return STEPS[: self._order]

    def next_steps(self) -> List["Step"]:
        return STEPS[self._order + 1 :]

    def __lt__(self, other) -> bool:
        if type(other) is type(self):
            return self._order < other._order

        return NotImplemented

    def __le__(self, other) -> bool:
        if type(other) is type(self):
            return self._order <= other._order

        return NotImplemented

    def __eq__(self, other) -> bool:
        if type(other) is type(self):
            return self.name == other.name

        return NotImplemented

    def __gt__(self, other) -> bool:
        if type(other) is type(self):
            return self._order > other._order

        return NotImplemented

    def __ge__(self, other) -> bool:
        if type(other) is type(self):
            return self._order >= other._order

        return NotImplemented

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "Step({!r})".format(self.name)


# Step names and order are now maintained in a single place: right here. If
# another step is needed, add it here, and insert it into STEPS in the proper
# order.
PULL = Step("pull", False)
BUILD = Step("build", False)
STAGE = Step("stage", True)
PRIME = Step("prime", True)

STEPS = [PULL, BUILD, STAGE, PRIME]


def next_step(step):
    """Get the next step of the lifecycle

    :param Step step: The current step. If None, the next step is the first
                      step in the lifecycle.
    :return: The next step in the lifecycle
    :rtype: Step
    """
    if step:
        return step.next_step()
    else:
        return STEPS[0]


def get_step_by_name(step_name):
    """Get the lifecycle step that has the given name.

    :param str step_name: Name of the step in question.
    :return: The Step in the lifecycle that has the given name.
    :rtype: Step
    :raises: errors.InvalidStepError if there is no step with the given name.
    """
    if step_name:
        for step in STEPS:
            if step.name == step_name:
                return step
        raise errors.InvalidStepError(step_name)
    else:
        return STEPS[0]


def get_dependency_prerequisite_step(step):
    if step <= STAGE:
        return STAGE
    else:
        return step


def dirty_step_if_dependency_changes(changed_step):
    if changed_step <= STAGE:
        return STEPS[0]
    else:
        return changed_step



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Sequence

from snapcraft.internal.os_release import OsRelease
from ._platform import _is_deb_based
from snapcraft.internal import errors

from typing import List


class RepoError(errors.SnapcraftError):
    pass


class NoNativeBackendError(RepoError):

    fmt = "Native builds aren't supported on {distro}."

    def __init__(self):
        try:
            distro = OsRelease().name()
        except errors.OsReleaseNameError:
            distro = "this system"
        super().__init__(distro=distro)


class CacheUpdateFailedError(RepoError):

    fmt = (
        "Failed to update the package cache: "
        "Some files could not be downloaded:{errors}"
        "Check that the sources on your host are configured correctly."
    )

    def __init__(self, errors: str) -> None:
        if errors:
            errors = "\n\n{}\n\n".format(errors.replace(", ", "\n"))
        else:
            errors = " "
        super().__init__(errors=errors)


class FileProviderNotFound(RepoError):

    fmt = "{file_path} is not provided by any package."

    def __init__(self, *, file_path: str) -> None:
        super().__init__(file_path=file_path)


class BuildPackageNotFoundError(RepoError):

    fmt = "Could not find a required package in 'build-packages': {package}"

    def __init__(self, package):
        super().__init__(package=package)


class BuildPackagesNotInstalledError(RepoError):

    fmt = "Could not install all requested build packages: {packages}"

    def __init__(self, *, packages: List[str]) -> None:
        super().__init__(packages=" ".join(packages))


class PackageFetchError(RepoError):

    fmt = "Package fetch error: {message}"

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class PackageBrokenError(RepoError):

    fmt = "The package {package} has unmet dependencies: {deps}"

    def __init__(self, package: str, deps: List[str]) -> None:
        super().__init__(package=package, deps=" ".join(deps))


class PackageNotFoundError(RepoError):
    @property
    def message(self):
        message = "The package {!r} was not found.".format(self.package_name)
        # If the package was multiarch, try to help.
        distro = OsRelease().id()
        if _is_deb_based(distro) and ":" in self.package_name:
            (name, arch) = self.package_name.split(":", 2)
            if arch:
                message += (
                    "\nYou may need to add support for this architecture with "
                    "'dpkg --add-architecture {}'.".format(arch)
                )
        return message

    def __init__(self, package_name):
        self.package_name = package_name

    def __str__(self):
        return self.message


class UnpackError(RepoError):

    fmt = "Error while provisioning {package!r}"

    def __init__(self, package):
        super().__init__(package=package)


class SnapUnavailableError(RepoError):

    fmt = (
        "Failed to install or refresh a snap: {snap_name!r} does not exist "
        "or is not available on the desired channel {snap_channel!r}. "
        "Use `snap info {snap_name}` to get a list of channels the "
        "snap is available on."
    )

    def __init__(self, *, snap_name: str, snap_channel: str) -> None:
        super().__init__(snap_name=snap_name, snap_channel=snap_channel)


class SnapFindError(RepoError):

    fmt = (
        "Could not find the snap {snap_name!r} installed on this host.\n"
        "Install the snap and try again."
    )

    def __init__(self, *, snap_name):
        super().__init__(snap_name=snap_name)


class SnapInstallError(RepoError):

    fmt = "Error while installing snap {snap_name!r} from channel {snap_channel!r}"

    def __init__(self, *, snap_name, snap_channel):
        super().__init__(snap_name=snap_name, snap_channel=snap_channel)


class SnapDownloadError(RepoError):

    fmt = "Error while downloading snap {snap_name!r} from channel {snap_channel!r}"

    def __init__(self, *, snap_name, snap_channel):
        super().__init__(snap_name=snap_name, snap_channel=snap_channel)


class SnapGetAssertionError(RepoError):

    fmt = (
        "Error while retrieving assertion with parameters "
        "{assertion_params!r}\n"
        "Verify the assertion exists and try again."
    )

    def __init__(self, *, assertion_params: Sequence[str]) -> None:
        super().__init__(assertion_params=assertion_params)


class SnapRefreshError(RepoError):

    fmt = "Error while refreshing snap {snap_name!r} to channel {snap_channel!r}"

    def __init__(self, *, snap_name, snap_channel):
        super().__init__(snap_name=snap_name, snap_channel=snap_channel)


class SnapdConnectionError(RepoError):

    fmt = (
        "Failed to get information for snap {snap_name!r}: "
        "could not connect to {url!r}."
    )

    def __init__(self, snap_name: str, url: str) -> None:
        super().__init__(snap_name=snap_name, url=url)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
from typing import Dict

from ._client import Client
from .info import SnapInfo

from . import errors
from . import logger, _macaroon_auth
from . import constants


class SnapIndexClient(Client):
    """The Click Package Index knows everything about existing snaps.
    https://wiki.ubuntu.com/AppStore/Interfaces/ClickPackageIndex is the
    canonical reference.
    """

    def __init__(self, conf):
        """Initialize the SnapIndexClient object.

        :param config conf: Configuration details for the client.
        :type config: snapcraft.config.Config
        """
        super().__init__(
            conf,
            os.environ.get(
                "UBUNTU_STORE_SEARCH_ROOT_URL", constants.UBUNTU_STORE_SEARCH_ROOT_URL
            ),
        )

    def get_default_headers(self, api="v2"):
        """Return default headers for CPI requests.
        Tries to build an 'Authorization' header with local credentials
        if they are available.
        Also pin specific branded store if `SNAPCRAFT_UBUNTU_STORE`
        environment is set.
        """
        headers = {}

        with contextlib.suppress(errors.InvalidCredentialsError):
            headers["Authorization"] = _macaroon_auth(self.conf)

        branded_store = os.getenv("SNAPCRAFT_UBUNTU_STORE")
        if branded_store:
            if api == "v2":
                headers["Snap-Device-Store"] = branded_store
            elif api == "v1":
                headers["X-Ubuntu-Store"] = branded_store
            else:
                logger.warning("Incorrect API version passed: {!r}.".format(api))

        return headers

    def get_info(self, snap_name: str, *, arch: str = None) -> SnapInfo:
        """Get the information for the specified snap.

        :param str snap_name: Name of the snap.
        :param str arch: Architecture of the snap (none by default).

        :return information for the snap.
        :rtype: SnapInfo
        """
        headers = self.get_default_headers()
        headers.update(
            {
                "Accept": "application/json",
                "Snap-Device-Series": constants.DEFAULT_SERIES,
            }
        )

        params = dict()
        params[
            "fields"
        ] = "channel-map,snap-id,name,publisher,confinement,revision,download"
        if arch is not None:
            params["architecture"] = arch
        logger.debug("Getting information for {}".format(snap_name))
        url = "v2/snaps/info/{}".format(snap_name)
        resp = self.get(url, headers=headers, params=params)
        if resp.status_code == 404:
            raise errors.SnapNotFoundError(snap_name, arch)
        resp.raise_for_status()

        return SnapInfo(resp.json())

    def get_assertion(
        self, assertion_type: str, snap_id: str
    ) -> Dict[str, Dict[str, str]]:
        """Get the assertion for the specified snap.

        :param str assertion_type: The type of the assertion.
        :param str snap_id: The ID of the snap.

        :return Assertion for the snap.
        """
        headers = self.get_default_headers(api="v1")
        logger.debug("Getting snap-declaration for {}".format(snap_id))
        url = "/api/v1/snaps/assertions/{}/{}/{}".format(
            assertion_type, constants.DEFAULT_SERIES, snap_id
        )
        response = self.get(url, headers=headers)
        if response.status_code != 200:
            raise errors.SnapNotFoundError(snap_id, series=constants.DEFAULT_SERIES)
        return response.json()

    def get(self, url, headers=None, params=None, stream=False):
        """Perform a GET request with the given arguments.

        :param str url: URL to send the request.
        :param dict headers: Headers to be sent along with the request.
        :param dict params: Query parameters to be sent along with
        the request.
        :param bool stream: Determines if the request shouldn't be
        automatically closed (true by default).

        :return Response of the request.
        """
        if headers is None:
            headers = self.get_default_headers()
        response = self.request(
            "GET", url, stream=stream, headers=headers, params=params
        )
        return response



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import os
import shutil
import sys
from textwrap import dedent
from typing import Optional, Sequence
from typing import Any, Dict

from xdg import BaseDirectory

from . import errors
from ._snap import SnapInjector
from snapcraft.internal import common, steps
from snapcraft import yaml_utils


def _get_platform() -> str:
    return sys.platform


def _get_tzdata(timezone_filepath=os.path.join(os.path.sep, "etc", "timezone")) -> str:
    """Return the host's timezone from timezone_filepath or Etc/UTC on error."""
    try:
        with open(timezone_filepath) as timezone_file:
            timezone = timezone_file.read().strip()
    except FileNotFoundError:
        timezone = "Etc/UTC"

    return timezone


# cloud-init's timezone keyword is not used as it requires tzdata to be installed
# and the images used may not have it preinstalled.
_CLOUD_USER_DATA_TMPL = dedent(
    """\
    #cloud-config
    manage_etc_hosts: true
    package_update: false
    growpart:
        mode: growpart
        devices: ["/"]
        ignore_growroot_disabled: false
    runcmd:
    - ["ln", "-s", "../usr/share/zoneinfo/{timezone}", "/etc/localtime"]
    write_files:
        - path: /root/.bashrc
          permissions: 0644
          content: |
            export SNAPCRAFT_BUILD_ENVIRONMENT=managed-host
            export PS1="\h \$(/bin/_snapcraft_prompt)# "
            export PATH=/snap/bin:$PATH
        - path: /bin/_snapcraft_prompt
          permissions: 0755
          content: |
            #!/bin/bash
            if [[ "$PWD" =~ ^$HOME.* ]]; then
                path="${{PWD/#$HOME/\ ..}}"
                if [[ "$path" == " .." ]]; then
                    ps1=""
                else
                    ps1="$path"
                fi
            else
                ps1="$PWD"
            fi
            echo -n $ps1
    """  # noqa: W605
)


class Provider(abc.ABC):

    _INSTANCE_PROJECT_DIR = "~/project"

    def __init__(self, *, project, echoer, is_ephemeral: bool = False) -> None:
        self.project = project
        self.echoer = echoer
        self._is_ephemeral = is_ephemeral

        self.instance_name = "snapcraft-{}".format(project.info.name)

        if project.info.version:
            self.snap_filename = "{}_{}_{}.snap".format(
                project.info.name, project.info.version, project.deb_arch
            )
        else:
            self.snap_filename = "{}_{}.snap".format(
                project.info.name, project.deb_arch
            )

        self.provider_project_dir = os.path.join(
            BaseDirectory.save_data_path("snapcraft"),
            "projects",
            project.info.name,
            self._get_provider_name(),
        )

    def __enter__(self):
        try:
            self.create()
        except errors.ProviderBaseError:
            # Destroy is idempotent
            self.destroy()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.destroy()

    @classmethod
    @abc.abstractclassmethod
    def ensure_provider(cls) -> None:
        """Necessary steps to ensure the provider is correctly setup."""

    @classmethod
    @abc.abstractclassmethod
    def setup_provider(cls, *, echoer) -> None:
        """Necessary steps to install the provider on the host."""

    @classmethod
    @abc.abstractclassmethod
    def _get_provider_name(cls) -> str:
        """Return the provider name."""

    @classmethod
    @abc.abstractclassmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        """Return whether the provider can install snaps from the host."""

    @abc.abstractmethod
    def create(self) -> None:
        """Provider steps needed to create a fully functioning environment."""

    @abc.abstractmethod
    def destroy(self) -> None:
        """Provider steps needed to ensure the instance is destroyed.

        This method should be safe to call multiple times and do nothing
        if the instance to destroy is already destroyed.
        """

    @abc.abstractmethod
    def _run(self, command: Sequence[str]) -> Optional[bytes]:
        """Run a command on the instance."""

    @abc.abstractmethod
    def _launch(self):
        """Launch the instance."""

    @abc.abstractmethod
    def _start(self):
        """Start an existing the instance."""

    @abc.abstractmethod
    def _push_file(self, *, source: str, destination: str) -> None:
        """Push a file into the instance."""

    @abc.abstractmethod
    def mount_project(self) -> None:
        """Provider steps needed to make the project available to the instance.
        """

    @abc.abstractmethod
    def _mount_prime_directory(self) -> bool:
        """Mount the host prime directory into the provider.

        :returns: True if the prime directory was already mounted.
        """

    def expose_prime(self) -> None:
        """Provider steps needed to expose the prime directory to the host.
        """
        os.makedirs(self.project.prime_dir, exist_ok=True)
        if not self._mount_prime_directory():
            self._run(command=["snapcraft", "clean", "--unprime"])

    def execute_step(self, step: steps.Step) -> None:
        self._run(command=["snapcraft", step.name])

    def clean(self, part_names: Sequence[str]) -> None:
        self._run(command=["snapcraft", "clean"] + list(part_names))

    def pack_project(self, *, output: Optional[str] = None) -> None:
        command = ["snapcraft", "snap"]
        if output:
            command.extend(["--output", output])
        self._run(command=command)

    def clean_project(self) -> bool:
        try:
            shutil.rmtree(self.provider_project_dir)
            return True
        except FileNotFoundError:
            return False

    @abc.abstractmethod
    def pull_file(self, name: str, destination: str, delete: bool = False) -> None:
        """
        Provider steps needed to retrieve a file from the instance, optionally
        deleting the source file after a successful retrieval.

        :param name: the remote filename.
        :type name: str
        :param destination: the local filename.
        :type destination: str
        :param delete: whether the file should be deleted.
        :type delete: bool
        """

    @abc.abstractmethod
    def shell(self) -> None:
        """Provider steps to provide a shell into the instance."""

    def launch_instance(self) -> None:
        # Check provider base and clean project if base has changed.
        if os.path.exists(self.provider_project_dir):
            self._ensure_base()

        try:
            # An ProviderStartError exception here means we need to create
            self._start()
        except errors.ProviderInstanceNotFoundError:
            # If starting failed, we need to make sure our data directory is
            # clean
            if os.path.exists(self.provider_project_dir):
                shutil.rmtree(self.provider_project_dir)
            os.makedirs(self.provider_project_dir)
            # then launch
            self._launch()
            # We need to setup snapcraft now to be able to refresh
            self._setup_snapcraft()
            # and do first boot related things
            self._run(["snapcraft", "refresh"])
        else:
            # We always setup snapcraft after a start to bring it up to speed with
            # what is on the host
            self._setup_snapcraft()

    def _ensure_base(self) -> None:
        info = self._load_info()
        provider_base = info["base"] if "base" in info else None
        if self._base_has_changed(self.project.info.get_build_base(), provider_base):
            self.echoer.warning(
                "Project base changed from {!r} to {!r}, cleaning build instance.".format(
                    provider_base, self.project.info.get_build_base()
                )
            )
            self.clean_project()

    def _setup_snapcraft(self) -> None:
        self._save_info(base=self.project.info.get_build_base())

        registry_filepath = os.path.join(
            self.provider_project_dir, "snap-registry.yaml"
        )

        # We do not want to inject from the host if not running from the snap
        # or if the provider cannot handle snap mounts.
        # This latter problem should go away when API for retrieving snaps
        # through snapd is generally available.
        if self._get_is_snap_injection_capable():
            inject_from_host = common.is_snap()
        else:
            inject_from_host = False

        snap_injector = SnapInjector(
            registry_filepath=registry_filepath,
            snap_arch=self.project.deb_arch,
            runner=self._run,
            file_pusher=self._push_file,
            inject_from_host=inject_from_host,
        )
        # Inject snapcraft
        snap_injector.add(snap_name="core")
        snap_injector.add(snap_name="snapcraft")

        # This build can be driven from a non snappy enabled system, so we may
        # find ourself in a situation where the base is not set like on OSX or
        # Windows.
        if self.project.info.get_build_base() is not None:
            snap_injector.add(snap_name=self.project.info.get_build_base())

        snap_injector.apply()

    def _get_cloud_user_data_string(self, timezone=_get_tzdata()) -> str:
        return _CLOUD_USER_DATA_TMPL.format(timezone=timezone)

    def _get_cloud_user_data(self, timezone=_get_tzdata()) -> str:
        cloud_user_data_filepath = os.path.join(
            self.provider_project_dir, "user-data.yaml"
        )
        if os.path.exists(cloud_user_data_filepath):
            return cloud_user_data_filepath

        user_data = self._get_cloud_user_data_string(timezone=timezone)

        with open(cloud_user_data_filepath, "w") as cloud_user_data_file:
            print(user_data, file=cloud_user_data_file, end="")

        return cloud_user_data_filepath

    def _base_has_changed(self, base: str, provider_base: str) -> bool:
        # Make it backwards compatible with instances without project info
        if base == "core18" and provider_base is None:
            return False
        elif base != provider_base:
            return True

        return False

    def _load_info(self) -> Dict[str, Any]:
        filepath = os.path.join(self.provider_project_dir, "project-info.yaml")
        if not os.path.exists(filepath):
            return dict()

        with open(filepath) as info_file:
            return yaml_utils.load(info_file)

    def _save_info(self, **data: Dict[str, Any]) -> None:
        filepath = os.path.join(self.provider_project_dir, "project-info.yaml")

        dirpath = os.path.dirname(filepath)
        if dirpath:
            os.makedirs(dirpath, exist_ok=True)

        with open(filepath, "w") as info_file:
            yaml_utils.dump(data, stream=info_file)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The python plugin can be used for python 2 or 3 based parts.

It can be used for python projects where you would want to do:

    - import python modules with a requirements.txt
    - build a python project that has a setup.py
    - install packages straight from pip

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - requirements:
      (list of strings)
      List of paths to requirements files.
    - constraints:
      (list of strings)
      List of paths to constraint files.
    - process-dependency-links:
      (bool; default: false)
      Enable the processing of dependency links in pip, which allow one
      project to provide places to look for another project
    - python-packages:
      (list)
      A list of dependencies to get from PyPI
    - python-version:
      (string; default: python3)
      The python version to use. Valid options are: python2 and python3

If the plugin finds a python interpreter with a basename that matches
`python-version` in the <stage> directory on the following fixed path:
`<stage-dir>/usr/bin/<python-interpreter>` then this interpreter would
be preferred instead and no interpreter would be brought in through
`stage-packages` mechanisms.
"""

import collections
import contextlib
import os
import re
from shutil import which
from textwrap import dedent
from typing import List, Set

import requests

import snapcraft
from snapcraft.common import isurl
from snapcraft.internal import errors, mangling
from snapcraft.internal.errors import SnapcraftPluginCommandError
from snapcraft.plugins import _python


class UnsupportedPythonVersionError(snapcraft.internal.errors.SnapcraftError):

    fmt = "Unsupported python version: {python_version!r}"


class SnapcraftPluginPythonFileMissing(snapcraft.internal.errors.SnapcraftError):

    fmt = (
        "Failed to find the referred {plugin_property} file at the given "
        "path: {plugin_property_value!r}.\n"
        "Check the property and ensure the file exists."
    )

    def __init__(self, *, plugin_property, plugin_property_value):
        super().__init__(
            plugin_property=plugin_property, plugin_property_value=plugin_property_value
        )


class PythonPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["requirements"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["constraints"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["python-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["process-dependency-links"] = {
            "type": "boolean",
            "default": False,
        }
        schema["properties"]["python-version"] = {
            "type": "string",
            "default": "python3",
            "enum": ["python2", "python3"],
        }
        schema["anyOf"] = [{"required": ["source"]}, {"required": ["python-packages"]}]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return [
            "requirements",
            "constraints",
            "python-packages",
            "process-dependency-links",
            "python-version",
        ]

    @property
    def plugin_stage_packages(self):
        if self.options.python_version == "python2":
            python_base = "python"
        elif self.options.python_version == "python3":
            python_base = "python3"

        if self.project.info.get_build_base() in ("core", "core16", "core18"):
            stage_packages = [python_base]
        else:
            stage_packages = []

        if self.project.info.get_build_base() == "core18" and python_base == "python3":
            stage_packages.append("{}-distutils".format(python_base))

        return stage_packages

    # ignore mypy error: Read-only property cannot override read-write property
    @property  # type: ignore
    def stage_packages(self):
        try:
            _python.get_python_command(
                self._python_major_version,
                stage_dir=self.project.stage_dir,
                install_dir=self.installdir,
            )
        except _python.errors.MissingPythonCommandError:
            return super().stage_packages + self.plugin_stage_packages
        else:
            return super().stage_packages

    @property
    def _pip(self):
        if not self.__pip:
            self.__pip = _python.Pip(
                python_major_version=self._python_major_version,
                part_dir=self.partdir,
                install_dir=self.installdir,
                stage_dir=self.project.stage_dir,
            )
        return self.__pip

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_base_tools(project.info.get_build_base())

        self._manifest = collections.OrderedDict()

        # Pip requires only the major version of python rather than the command
        # name like our option requires.
        match = re.match(r"python(?P<major_version>\d).*", self.options.python_version)
        if not match:
            raise UnsupportedPythonVersionError(
                python_version=self.options.python_version
            )

        self._python_major_version = match.group("major_version")
        self.__pip = None

    def _setup_base_tools(self, base):
        # NOTE: stage-packages are lazily loaded.
        if base in ("core", "core16", "core18"):
            if self.options.python_version == "python3":
                self.build_packages.extend(
                    [
                        "python3-dev",
                        "python3-pip",
                        "python3-pkg-resources",
                        "python3-setuptools",
                    ]
                )
            elif self.options.python_version == "python2":
                self.build_packages.extend(
                    [
                        "python-dev",
                        "python-pip",
                        "python-pkg-resources",
                        "python-setuptools",
                    ]
                )
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def pull(self):
        super().pull()

        self._pip.setup()

        with simple_env_bzr(os.path.join(self.installdir, "bin")):
            # Download this project, using its setup.py if present. This will
            # also download any python-packages requested.
            self._download_project()

    def clean_pull(self):
        super().clean_pull()
        self._pip.clean_packages()

    def build(self):
        super().build()

        with simple_env_bzr(os.path.join(self.installdir, "bin")):
            # Install the packages that have already been downloaded
            installed_pipy_packages = self._install_project()

        requirements = self._get_list_of_packages_from_property(
            self.options.requirements
        )
        if requirements:
            self._manifest["requirements-contents"] = requirements

        constraints = self._get_list_of_packages_from_property(self.options.constraints)
        if constraints:
            self._manifest["constraints-contents"] = constraints

        self._manifest["python-packages"] = [
            "{}={}".format(name, installed_pipy_packages[name])
            for name in installed_pipy_packages
        ]

        _python.generate_sitecustomize(
            self._python_major_version,
            stage_dir=self.project.stage_dir,
            install_dir=self.installdir,
        )

    def _find_file(self, *, filename: str) -> str:
        # source-subdir defaults to ''
        for basepath in [self.builddir, self.sourcedir]:
            if basepath == self.sourcedir:
                # This is overwritten in the base plugin
                # TODO add consistency
                source_subdir = self.options.source_subdir
            else:
                source_subdir = ""
            filepath = os.path.join(basepath, source_subdir, filename)
            if os.path.exists(filepath):
                return filepath

        return None

    def _get_setup_py_dir(self):
        setup_py_dir = None
        setup_py_path = self._find_file(filename="setup.py")
        if setup_py_path:
            setup_py_dir = os.path.dirname(setup_py_path)

        return setup_py_dir

    def _get_list_of_packages_from_property(self, property_list: Set[str]) -> List[str]:
        """Return a sorted list of all packages found in property."""
        package_list = list()  # type: List[str]
        for entry in property_list:
            contents = self._get_file_contents(entry)
            package_list.extend(contents.splitlines())
        return package_list

    def _get_normalized_property_set(
        self, property_name, property_list: List[str]
    ) -> Set[str]:
        """Return a normalized set from a requirements or constraints list."""
        normalized = set()  # type: Set[str]
        for entry in property_list:
            if isurl(entry):
                normalized.add(entry)
            else:
                entry_file = self._find_file(filename=entry)
                if not entry_file:
                    raise SnapcraftPluginPythonFileMissing(
                        plugin_property=property_name, plugin_property_value=entry
                    )
                normalized.add(entry_file)

        return normalized

    def _install_wheels(self, wheels):
        installed = self._pip.list()
        wheel_names = [os.path.basename(w).split("-")[0] for w in wheels]

        # we want to avoid installing what is already provided in
        # stage-packages
        need_install = [k for k in wheel_names if k not in installed]
        self._pip.install(
            need_install,
            upgrade=True,
            install_deps=False,
            process_dependency_links=self.options.process_dependency_links,
        )

    def _download_project(self):
        constraints = self._get_normalized_property_set(
            "constraints", self.options.constraints
        )
        requirements = self._get_normalized_property_set(
            "requirements", self.options.requirements
        )

        self._pip.download(
            self.options.python_packages,
            setup_py_dir=None,
            constraints=constraints,
            requirements=requirements,
            process_dependency_links=self.options.process_dependency_links,
        )

    def _install_project(self):
        setup_py_dir = self._get_setup_py_dir()
        constraints = self._get_normalized_property_set(
            "constraints", self.options.constraints
        )
        requirements = self._get_normalized_property_set(
            "requirements", self.options.requirements
        )

        # setup.py is handled in a different step as some projects may
        # need to satisfy dependencies for setup.py to be parsed.
        wheels = self._pip.wheel(
            self.options.python_packages,
            setup_py_dir=None,
            constraints=constraints,
            requirements=requirements,
            process_dependency_links=self.options.process_dependency_links,
        )

        if wheels:
            self._install_wheels(wheels)

        if setup_py_dir is not None:
            self._pip.download(
                [],
                setup_py_dir=setup_py_dir,
                constraints=constraints,
                requirements=set(),
                process_dependency_links=self.options.process_dependency_links,
            )
            wheels = self._pip.wheel(
                [],
                setup_py_dir=setup_py_dir,
                constraints=constraints,
                requirements=set(),
                process_dependency_links=self.options.process_dependency_links,
            )

            if wheels:
                self._install_wheels(wheels)

            setup_py_path = os.path.join(setup_py_dir, "setup.py")
            if os.path.exists(setup_py_path):
                # pbr and others don't work using `pip install .`
                # LP: #1670852
                # There is also a chance that this setup.py is distutils based
                # in which case we will rely on the `pip install .` ran before
                #  this.
                with contextlib.suppress(SnapcraftPluginCommandError):
                    self._setup_tools_install(setup_py_path)

        return self._pip.list()

    def _setup_tools_install(self, setup_file):
        command = [
            _python.get_python_command(
                self._python_major_version,
                stage_dir=self.project.stage_dir,
                install_dir=self.installdir,
            ),
            os.path.basename(setup_file),
            "--no-user-cfg",
            "install",
            "--single-version-externally-managed",
            "--user",
            "--record",
            "install.txt",
        ]
        self.run(command, env=self._pip.env(), cwd=os.path.dirname(setup_file))

        # Fix all shebangs to use the in-snap python. The stuff installed from
        # pip has already been fixed, but anything done in this step has not.
        mangling.rewrite_python_shebangs(self.installdir)

    def _get_file_contents(self, path):
        if isurl(path):
            return requests.get(path).text
        else:
            file_path = os.path.join(self.sourcedir, path)
            with open(file_path) as _file:
                return _file.read()

    def get_manifest(self):
        return self._manifest

    def snap_fileset(self):
        fileset = super().snap_fileset()
        fileset.append("-bin/pip")
        fileset.append("-bin/pip2")
        fileset.append("-bin/pip3")
        fileset.append("-bin/pip2.7")
        fileset.append("-bin/pip3.*")
        fileset.append("-bin/easy_install*")
        fileset.append("-bin/wheel")
        # Holds all the .pyc files. It is a major cause of inter part
        # conflict.
        fileset.append("-**/__pycache__")
        fileset.append("-**/*.pyc")
        # The RECORD files include hashes useful when uninstalling packages.
        # In the snap they will cause conflicts when more than one part uses
        # the python plugin.
        fileset.append("-lib/python*/site-packages/*/RECORD")
        return fileset


@contextlib.contextmanager
def simple_env_bzr(bin_dir):
    """Create an appropriate environment to run bzr.

       The python plugin sets up PYTHONUSERBASE and PYTHONHOME which
       conflicts with bzr when using python3 as those two environment
       variables will make bzr look for modules in the wrong location.
       """
    os.makedirs(bin_dir, exist_ok=True)
    bzr_bin = os.path.join(bin_dir, "bzr")
    real_bzr_bin = which("bzr")
    if real_bzr_bin:
        exec_line = 'exec {} "$@"'.format(real_bzr_bin)
    else:
        exec_line = "echo bzr needs to be in PATH; exit 1"
    with open(bzr_bin, "w") as f:
        f.write(
            dedent(
                """#!/bin/sh
               unset PYTHONUSERBASE
               unset PYTHONHOME
               {}
            """.format(
                    exec_line
                )
            )
        )
    os.chmod(bzr_bin, 0o777)
    try:
        yield
    finally:
        os.remove(bzr_bin)
        if not os.listdir(bin_dir):
            os.rmdir(bin_dir)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The catkin plugin is useful for building ROS parts.

The rosdistro used depends upon the base of the snap:

  - core: Uses Kinetic
  - core16: Uses Kinetic
  - core18: Uses Melodic

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - catkin-packages:
      (list of strings)
      List of catkin packages to build. If not specified, all packages in the
      workspace will be built. If set to an empty list ([]), no packages will
      be built.
    - source-space:
      (string)
      The source space containing Catkin packages. By default this is 'src'.
    - include-roscore:
      (boolean)
      Whether or not to include roscore with the part. Defaults to true.
    - rosinstall-files:
      (list of strings)
      List of rosinstall files to merge while pulling. Paths are relative to
      the source.
    - recursive-rosinstall:
      (boolean)
      Whether or not to recursively merge/update rosinstall files from fetched
      sources. Will continue until all rosinstall files have been merged.
      Defaults to false.
    - catkin-cmake-args:
      (list of strings)
      Configure flags to pass onto the cmake invocation from catkin.
    - underlay:
      (object)
      Used to inform Snapcraft that this snap isn't standalone, and is actually
      overlaying a workspace from another snap via content sharing. Made up of
      two properties:
      - build-path:
        (string)
        Build-time path to existing workspace to underlay the one being built,
        for example '$SNAPCRAFT_STAGE/opt/ros/kinetic'.
      - run-path:
        (string)
        Run-time path of the underlay workspace (e.g. a subdirectory of the
        content interface's 'target' attribute.)
    - catkin-ros-master-uri:
      (string)
      The URI to ros master setting the env variable ROS_MASTER_URI. Defaults
      to http://localhost:11311.
"""

import contextlib
import glob
import os
import tempfile
import logging
import re
import shutil
import subprocess
import textwrap
from typing import List

import snapcraft
from snapcraft.plugins import _ros
from snapcraft.plugins import _python
from snapcraft import common, file_utils, formatting_utils, repo
from snapcraft.internal import errors, mangling

logger = logging.getLogger(__name__)

# Map bases to ROS releases
_BASE_TO_ROS_RELEASE_MAP = {"core": "kinetic", "core16": "kinetic", "core18": "melodic"}

# Map bases to Ubuntu releases
_BASE_TO_UBUNTU_RELEASE_MAP = {"core": "xenial", "core16": "xenial", "core18": "bionic"}

_SUPPORTED_DEPENDENCY_TYPES = {"apt", "pip"}

_ROS_KEYRING_PATH = os.path.join(snapcraft.internal.common.get_keyringsdir(), "ros.gpg")


class CatkinInvalidSystemDependencyError(errors.SnapcraftError):
    fmt = (
        "Package {dependency!r} isn't a valid system dependency. Did you "
        "forget to add it to catkin-packages? If not, add the Ubuntu package "
        "containing it to stage-packages until you can get it into the rosdep "
        "database."
    )

    def __init__(self, dependency):
        super().__init__(dependency=dependency)


class CatkinUnsupportedDependencyTypeError(errors.SnapcraftError):
    fmt = (
        "Package {dependency!r} resolved to an unsupported type of "
        "dependency: {dependency_type!r}."
    )

    def __init__(self, dependency_type, dependency):
        super().__init__(dependency_type=dependency_type, dependency=dependency)


class CatkinWorkspaceIsRootError(errors.SnapcraftError):
    fmt = "source-space cannot be the root of the Catkin workspace; use a subdirectory."


class CatkinCannotResolveRoscoreError(errors.SnapcraftError):
    fmt = "Failed to determine system dependency for roscore."


class CatkinAptDependencyFetchError(errors.SnapcraftError):
    fmt = "Failed to fetch apt dependencies: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class CatkinNoHighestVersionPathError(errors.SnapcraftError):
    fmt = "Failed to determine highest path in {path!r}: nothing found."

    def __init__(self, path):
        super().__init__(path=path)


class CatkinGccVersionError(errors.SnapcraftError):
    fmt = "Failed to determine gcc version: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class CatkinPackagePathNotFoundError(errors.SnapcraftError):
    fmt = "Failed to find package path: {path!r}"

    def __init__(self, path):
        super().__init__(path=path)


class CatkinPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["catkin-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
        }
        schema["properties"]["source-space"] = {"type": "string", "default": "src"}

        # The default is true since we expect most Catkin packages to be ROS
        # packages. The only reason one wouldn't want to include ROS in the
        # snap is if library snaps exist, which will still likely be the
        # minority.
        schema["properties"]["include-roscore"] = {"type": "boolean", "default": True}

        schema["properties"]["underlay"] = {
            "type": "object",
            "properties": {
                "build-path": {"type": "string"},
                "run-path": {"type": "string"},
            },
            "required": ["build-path", "run-path"],
        }

        schema["properties"]["rosinstall-files"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["recursive-rosinstall"] = {
            "type": "boolean",
            "default": False,
        }

        schema["properties"]["catkin-cmake-args"] = {
            "type": "array",
            "minitems": 1,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["catkin-ros-master-uri"] = {
            "type": "string",
            "default": "http://localhost:11311",
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return [
            "catkin-packages",
            "source-space",
            "include-roscore",
            "underlay",
            "rosinstall-files",
            "recursive-rosinstall",
        ]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["catkin-cmake-args"]

    @property
    def _pip(self):
        if not self.__pip:
            self.__pip = _python.Pip(
                python_major_version="2",  # ROS1 only supports python2
                part_dir=self.partdir,
                install_dir=self.installdir,
                stage_dir=self.project.stage_dir,
            )

        return self.__pip

    @property
    def PLUGIN_STAGE_SOURCES(self):
        ros_repo = "http://packages.ros.org/ros/ubuntu/"
        ubuntu_repo = "http://${prefix}.ubuntu.com/${suffix}/"
        security_repo = "http://${security}.ubuntu.com/${suffix}/"

        return textwrap.dedent(
            """
            deb {ros_repo} {codename} main
            deb {ubuntu_repo} {codename} main universe
            deb {ubuntu_repo} {codename}-updates main universe
            deb {ubuntu_repo} {codename}-security main universe
            deb {security_repo} {codename}-security main universe
            """.format(
                ros_repo=ros_repo,
                ubuntu_repo=ubuntu_repo,
                security_repo=security_repo,
                codename=_BASE_TO_UBUNTU_RELEASE_MAP[
                    self.project.info.get_build_base()
                ],
            )
        )

    @property
    def PLUGIN_STAGE_KEYRINGS(self):
        return [_ROS_KEYRING_PATH]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        base = self.project.info.get_build_base()
        if base not in ("core", "core16", "core18"):
            raise errors.PluginBaseError(part_name=self.name, base=base)

        self._rosdistro = _BASE_TO_ROS_RELEASE_MAP[base]

        self.build_packages.extend(["gcc", "g++", "libc6-dev", "make", "python-pip"])
        self.__pip = None

        # roslib is the base requiremet to actually create a workspace with
        # setup.sh and the necessary hooks.
        self.stage_packages.append("ros-{}-roslib".format(self._rosdistro))

        # Get a unique set of packages
        self.catkin_packages = None
        if options.catkin_packages is not None:
            self.catkin_packages = set(options.catkin_packages)
        self._rosdep_path = os.path.join(self.partdir, "rosdep")
        self._rospack_path = os.path.join(self.partdir, "rospack")
        self._catkin_path = os.path.join(self.partdir, "catkin")
        self._wstool_path = os.path.join(self.partdir, "wstool")

        # The path created via the `source` key (or a combination of `source`
        # and `source-subdir` keys) needs to point to a valid Catkin workspace
        # containing another subdirectory called the "source space." By
        # default, this is a directory named "src," but it can be remapped via
        # the `source-space` key. It's important that the source space is not
        # the root of the Catkin workspace, since Catkin won't work that way
        # and it'll create a circular link that causes rosdep to hang.
        if self.options.source_subdir:
            self._ros_package_path = os.path.join(
                self.sourcedir, self.options.source_subdir, self.options.source_space
            )
        else:
            self._ros_package_path = os.path.join(
                self.sourcedir, self.options.source_space
            )

        if os.path.abspath(self.sourcedir) == os.path.abspath(self._ros_package_path):
            raise CatkinWorkspaceIsRootError()

    def env(self, root):
        """Runtime environment for ROS binaries and services."""

        paths = common.get_library_paths(root, self.project.arch_triplet)
        ld_library_path = formatting_utils.combine_paths(
            paths, prepend="", separator=":"
        )

        env = [
            # This environment variable tells ROS nodes where to find ROS
            # master. It does not affect ROS master, however-- this is just the
            # URI.
            "ROS_MASTER_URI={}".format(self.options.catkin_ros_master_uri),
            # Various ROS tools (e.g. rospack, roscore) keep a cache or a log,
            # and use $ROS_HOME to determine where to put them.
            "ROS_HOME=${SNAP_USER_DATA:-/tmp}/ros",
            # FIXME: LP: #1576411 breaks ROS snaps on the desktop, so we'll
            # temporarily work around that bug by forcing the locale to
            # C.UTF-8.
            "LC_ALL=C.UTF-8",
            # The Snapcraft Core will ensure that we get a good LD_LIBRARY_PATH
            # overall, but it defines it after this function runs. Some ROS
            # tools will cause binaries to be run when we source the setup.sh,
            # below, so we need to have a sensible LD_LIBRARY_PATH before then.
            "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:{}".format(ld_library_path),
        ]

        # There's a chicken and egg problem here, everything run gets an
        # env built, even package installation, so the first runs for these
        # will likely fail.
        try:
            # The ROS packaging system tools (e.g. rospkg, etc.) don't go
            # into the ROS install path (/opt/ros/$distro), so we need the
            # PYTHONPATH to include the dist-packages in /usr/lib as well.
            #
            # Note: Empty segments in PYTHONPATH are interpreted as `.`, thus
            # adding the current working directory to the PYTHONPATH. That is
            # not desired in this situation, so take proper precautions when
            # expanding PYTHONPATH: only add it if it's not empty.
            env.append(
                "PYTHONPATH={}${{PYTHONPATH:+:$PYTHONPATH}}".format(
                    common.get_python2_path(root)
                )
            )
        except errors.SnapcraftEnvironmentError as e:
            logger.debug(e)

        # The setup.sh we source below requires the in-snap python. Here we
        # make sure it's in the PATH before it's run.
        env.append("PATH=$PATH:{}/usr/bin".format(root))

        if self.options.underlay:
            script = textwrap.dedent(
                """
                if [ -f {snapcraft_setup} ]; then
                    . {snapcraft_setup}
                fi
            """
            ).format(snapcraft_setup=os.path.join(self.rosdir, "snapcraft-setup.sh"))
        else:
            script = self._source_setup_sh(root, None)

        # Each of these lines is prepended with an `export` when the
        # environment is actually generated. In order to inject real shell code
        # we have to hack it in by appending it on the end of an item already
        # in the environment. FIXME: There should be a better way to do this.
        # LP: #1792034
        env[-1] = env[-1] + "\n\n" + script

        return env

    def pull(self):
        """Copy source into build directory and fetch dependencies.

        Catkin packages can specify their system dependencies in their
        package.xml. In order to support that, the Catkin packages are
        interrogated for their dependencies here. Since `stage-packages` are
        already installed by the time this function is run, the dependencies
        from the package.xml are pulled down explicitly.
        """

        super().pull()

        # There may be nothing contained within the source but a rosinstall
        # file. We need to use it to flesh out the workspace before continuing
        # with the pull.
        if self.options.rosinstall_files or self.options.recursive_rosinstall:
            wstool = _ros.wstool.Wstool(
                self._ros_package_path,
                self._wstool_path,
                self.PLUGIN_STAGE_SOURCES,
                self.PLUGIN_STAGE_KEYRINGS,
                self.project,
            )
            wstool.setup()

            source_path = self.sourcedir
            if self.options.source_subdir:
                source_path = os.path.join(self.sourcedir, self.options.source_subdir)

            # Recursively handling rosinstall files is a superset of handling
            # individual rosinstall files. If both are specified, the recursive
            # option will cover it.
            if self.options.recursive_rosinstall:
                _recursively_handle_rosinstall_files(wstool, source_path)
            else:
                # The rosinstall files in the YAML are relative to the part's
                # source. However, _handle_rosinstall_files requires absolute
                # paths.
                rosinstall_files = set()
                for rosinstall_file in self.options.rosinstall_files:
                    rosinstall_files.add(os.path.join(source_path, rosinstall_file))

                _handle_rosinstall_files(wstool, rosinstall_files)

        # Make sure the package path exists before continuing. We only care
        # about doing this if there are actually packages to build, which is
        # indicated both by self.catkin_packages being None as well as a
        # non-empty list.
        packages_to_build = (
            self.catkin_packages is None or len(self.catkin_packages) > 0
        )
        if packages_to_build and not os.path.exists(self._ros_package_path):
            raise CatkinPackagePathNotFoundError(self._ros_package_path)

        # Validate the underlay. Note that this validation can't happen in
        # __init__ as the underlay will probably only be valid once a
        # dependency has been staged.
        catkin = None
        underlay_build_path = None
        dependency_workspaces = [self.rosdir]
        if self.options.underlay:
            underlay_build_path = self.options.underlay["build-path"]
        if underlay_build_path:
            if not os.path.isdir(underlay_build_path):
                raise errors.SnapcraftEnvironmentError(
                    "Requested underlay ({!r}) does not point to a valid "
                    "directory".format(underlay_build_path)
                )

            if not os.path.isfile(os.path.join(underlay_build_path, "setup.sh")):
                raise errors.SnapcraftEnvironmentError(
                    "Requested underlay ({!r}) does not contain a "
                    "setup.sh".format(underlay_build_path)
                )

            dependency_workspaces.append(underlay_build_path)
            self._generate_snapcraft_setup_sh(self.installdir, underlay_build_path)

        # Use catkin_find to discover dependencies already in the underlay
        catkin = _Catkin(
            self._rosdistro,
            dependency_workspaces,
            self._catkin_path,
            self.PLUGIN_STAGE_SOURCES,
            self.PLUGIN_STAGE_KEYRINGS,
            self.project,
        )
        catkin.setup()

        # Use rospack to obtain list of packages in workspace
        rospack = _ros.rospack.Rospack(
            ros_distro=self._rosdistro,
            ros_package_path=self._ros_package_path,
            rospack_path=self._rospack_path,
            ubuntu_sources=self.PLUGIN_STAGE_SOURCES,
            ubuntu_keyrings=self.PLUGIN_STAGE_KEYRINGS,
            project=self.project,
        )
        rospack.setup()

        # Use rosdep for dependency detection and resolution
        rosdep = _ros.rosdep.Rosdep(
            ros_distro=self._rosdistro,
            ros_package_path=self._ros_package_path,
            rosdep_path=self._rosdep_path,
            ubuntu_distro=_BASE_TO_UBUNTU_RELEASE_MAP[
                self.project.info.get_build_base()
            ],
            ubuntu_sources=self.PLUGIN_STAGE_SOURCES,
            ubuntu_keyrings=self.PLUGIN_STAGE_KEYRINGS,
            project=self.project,
        )
        rosdep.setup()

        self._setup_dependencies(rosdep, rospack, catkin)

    def _setup_dependencies(self, rosdep, rospack, catkin):
        # Parse the Catkin packages to pull out their system dependencies
        system_dependencies = _find_system_dependencies(
            self.catkin_packages, rosdep, rospack, catkin
        )

        # If the package requires roscore, resolve it into a system dependency
        # as well.
        if self.options.include_roscore:
            roscore = rosdep.resolve_dependency("ros_core")
            if roscore:
                for dependency_type, dependencies in roscore.items():
                    if dependency_type not in system_dependencies:
                        system_dependencies[dependency_type] = set()
                    system_dependencies[dependency_type] |= dependencies
            else:
                raise CatkinCannotResolveRoscoreError()

        # Pull down and install any apt dependencies that were discovered
        self._setup_apt_dependencies(system_dependencies.get("apt"))

        # Pull down and install any pip dependencies that were discovered
        self._setup_pip_dependencies(system_dependencies.get("pip"))

    def _setup_apt_dependencies(self, apt_dependencies):
        if apt_dependencies:
            ubuntudir = os.path.join(self.partdir, "ubuntu")
            os.makedirs(ubuntudir, exist_ok=True)

            logger.info("Preparing to fetch apt dependencies...")
            ubuntu = repo.Ubuntu(
                ubuntudir,
                sources=self.PLUGIN_STAGE_SOURCES,
                keyrings=self.PLUGIN_STAGE_KEYRINGS,
                project_options=self.project,
            )

            logger.info("Fetching apt dependencies...")
            try:
                ubuntu.get(apt_dependencies)
            except repo.errors.PackageNotFoundError as e:
                raise CatkinAptDependencyFetchError(e.message)

            logger.info("Installing apt dependencies...")
            ubuntu.unpack(self.installdir)

    def _setup_pip_dependencies(self, pip_dependencies):
        if pip_dependencies:
            self._pip.setup()

            logger.info("Fetching pip dependencies...")
            self._pip.download(pip_dependencies)

            logger.info("Installing pip dependencies...")
            self._pip.install(pip_dependencies)

    def clean_pull(self):
        super().clean_pull()

        # Remove the rosdep path, if any
        with contextlib.suppress(FileNotFoundError):
            shutil.rmtree(self._rosdep_path)

        # Remove the catkin path, if any
        with contextlib.suppress(FileNotFoundError):
            shutil.rmtree(self._catkin_path)

        # Clean pip packages, if any
        self._pip.clean_packages()

    def _source_setup_sh(self, root, underlay_path):
        rosdir = os.path.join(root, "opt", "ros", self._rosdistro)
        if underlay_path:
            source_script = textwrap.dedent(
                """
                if [ -f {underlay_setup} ]; then
                    _CATKIN_SETUP_DIR={underlay} . {underlay_setup}
                    if [ -f {rosdir_setup} ]; then
                        set -- --extend
                        _CATKIN_SETUP_DIR={rosdir} . {rosdir_setup}
                    fi
                fi
            """
            ).format(
                underlay=underlay_path,
                underlay_setup=os.path.join(underlay_path, "setup.sh"),
                rosdir=rosdir,
                rosdir_setup=os.path.join(rosdir, "setup.sh"),
            )
        else:
            source_script = textwrap.dedent(
                """
                if [ -f {rosdir_setup} ]; then
                    _CATKIN_SETUP_DIR={rosdir} . {rosdir_setup}
                fi
            """
            ).format(rosdir=rosdir, rosdir_setup=os.path.join(rosdir, "setup.sh"))

        # We need to source ROS's setup.sh at this point. However, it accepts
        # arguments (thus will parse $@), and we really don't want it to, since
        # $@ in this context will be meant for the app being launched
        # (LP: #1660852). So we'll backup all args, source the setup.sh, then
        # restore all args for the wrapper's `exec` line.
        return textwrap.dedent(
            """
            # Shell quote arbitrary string by replacing every occurrence of '
            # with '\\'', then put ' at the beginning and end of the string.
            # Prepare yourself, fun regex ahead.
            quote()
            {{
                for i; do
                    printf %s\\\\n "$i" | sed "s/\'/\'\\\\\\\\\'\'/g;1s/^/\'/;\$s/\$/\' \\\\\\\\/"
                done
                echo " "
            }}

            BACKUP_ARGS=$(quote "$@")
            set --
            {}
            eval "set -- $BACKUP_ARGS"
        """  # noqa: W605
        ).format(
            source_script
        )  # noqa

    def _generate_snapcraft_setup_sh(self, root, underlay_path):
        script = self._source_setup_sh(root, underlay_path)
        os.makedirs(self.rosdir, exist_ok=True)
        with open(os.path.join(self.rosdir, "snapcraft-setup.sh"), "w") as f:
            f.write(script)

    @property
    def rosdir(self):
        return os.path.join(self.installdir, "opt", "ros", self._rosdistro)

    def _run_in_bash(self, commandlist, cwd=None, env=None):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write("set -e\n")
            f.write("exec {}\n".format(" ".join(commandlist)))
            f.flush()

            self.run(["/bin/bash", f.name], cwd=cwd, env=env)

    def build(self):
        """Build Catkin packages.

        This function runs some pre-build steps to prepare the sources for
        building in the Snapcraft environment, builds the packages via
        catkin_make_isolated, and finally runs some post-build clean steps
        to prepare the newly-minted install to be packaged as a .snap.
        """

        super().build()

        logger.info("Preparing to build Catkin packages...")
        self._prepare_build()

        logger.info("Building Catkin packages...")
        self._build_catkin_packages()

        logger.info("Cleaning up newly installed Catkin packages...")
        self._finish_build()

    def _prepare_build(self):
        self._use_in_snap_python()

        # Each Catkin package distributes .cmake files so they can be found via
        # find_package(). However, the Ubuntu packages pulled down as
        # dependencies contain .cmake files pointing to system paths (e.g.
        # /usr/lib, /usr/include, etc.). They need to be rewritten to point to
        # the install directory.
        def _new_path(path):
            if not path.startswith(self.installdir):
                # Not using os.path.join here as `path` is absolute.
                return self.installdir + path
            return path

        self._rewrite_cmake_paths(_new_path)

        # Also rewrite any occurrence of $SNAPCRAFT_STAGE to be our install
        # directory (this may be the case if stage-snaps were used).
        file_utils.replace_in_file(
            self.rosdir,
            re.compile(r".*Config.cmake$"),
            re.compile(r"\$ENV{SNAPCRAFT_STAGE}"),
            self.installdir,
        )

    def _rewrite_cmake_paths(self, new_path_callable):
        def _rewrite_paths(match):
            paths = match.group(1).strip().split(";")
            for i, path in enumerate(paths):
                # Offer the opportunity to rewrite this path if it's absolute.
                if os.path.isabs(path):
                    paths[i] = new_path_callable(path)

            return '"' + ";".join(paths) + '"'

        # Looking for any path-like string
        file_utils.replace_in_file(
            self.rosdir,
            re.compile(r".*Config.cmake$"),
            re.compile(r'"(.*?/.*?)"'),
            _rewrite_paths,
        )

    def _finish_build(self):
        self._use_in_snap_python()

        # We've finished the build, but we need to make sure we turn the cmake
        # files back into something that doesn't include our installdir. This
        # way it's usable from the staging area, and won't clash with the same
        # file coming from other parts.
        pattern = re.compile(r"^{}".format(self.installdir))

        def _new_path(path):
            return pattern.sub("$ENV{SNAPCRAFT_STAGE}", path)

        self._rewrite_cmake_paths(_new_path)

        # Replace the CMAKE_PREFIX_PATH in _setup_util.sh
        setup_util_file = os.path.join(self.rosdir, "_setup_util.py")
        if os.path.isfile(setup_util_file):
            with open(setup_util_file, "r+") as f:
                pattern = re.compile(r"CMAKE_PREFIX_PATH = '.*/opt/ros.*")
                replaced = pattern.sub("CMAKE_PREFIX_PATH = []", f.read())
                f.seek(0)
                f.truncate()
                f.write(replaced)

        # Set the _CATKIN_SETUP_DIR in setup.sh to a sensible default, removing
        # our installdir (this way it doesn't clash with a setup.sh coming
        # from another part).
        setup_sh_file = os.path.join(self.rosdir, "setup.sh")
        if os.path.isfile(setup_sh_file):
            with open(setup_sh_file, "r+") as f:
                pattern = re.compile(r"\${_CATKIN_SETUP_DIR:=.*}")
                replaced = pattern.sub(
                    "${{_CATKIN_SETUP_DIR:=$SNAP/opt/ros/{}}}".format(self._rosdistro),
                    f.read(),
                )
                f.seek(0)
                f.truncate()
                f.write(replaced)

        if self.options.underlay:
            underlay_run_path = self.options.underlay["run-path"]
            self._generate_snapcraft_setup_sh("$SNAP", underlay_run_path)

        # If pip dependencies were installed, generate a sitecustomize that
        # allows access to them.
        if self._pip.is_setup() and self._pip.list(user=True):
            _python.generate_sitecustomize(
                "2", stage_dir=self.project.stage_dir, install_dir=self.installdir
            )

    def _use_in_snap_python(self):
        # Fix all shebangs to use the in-snap python.
        mangling.rewrite_python_shebangs(self.installdir)

        # Also replace all the /usr/bin/python calls in etc/catkin/profile.d/
        # files with the in-snap python
        profile_d_path = os.path.join(self.rosdir, "etc", "catkin", "profile.d")
        file_utils.replace_in_file(
            profile_d_path, re.compile(r""), re.compile(r"/usr/bin/python"), r"python"
        )

    def _build_catkin_packages(self):
        # Nothing to do if no packages were specified
        if self.catkin_packages is not None and len(self.catkin_packages) == 0:
            return

        catkincmd = ["catkin_make_isolated"]

        # Install the package
        catkincmd.append("--install")

        if self.catkin_packages:
            # Specify the packages to be built
            catkincmd.append("--pkg")
            catkincmd.extend(self.catkin_packages)

        # Don't clutter the real ROS workspace-- use the Snapcraft build
        # directory
        catkincmd.extend(["--directory", self.builddir])

        # Account for a non-default source space by always specifying it
        catkincmd.extend(
            ["--source-space", os.path.join(self.builddir, self.options.source_space)]
        )

        # Specify that the package should be installed along with the rest of
        # the ROS distro.
        catkincmd.extend(["--install-space", self.rosdir])

        # Specify the number of workers
        catkincmd.append("-j{}".format(self.parallel_build_count))

        # All the arguments that follow are meant for CMake
        catkincmd.append("--cmake-args")

        build_type = "Release"
        if "debug" in self.options.build_attributes:
            build_type = "Debug"
        catkincmd.append("-DCMAKE_BUILD_TYPE={}".format(build_type))

        # Finally, add any cmake-args requested from the plugin options
        catkincmd.extend(self.options.catkin_cmake_args)

        # This command must run in bash due to a bug in Catkin that causes it
        # to explode if there are spaces in the cmake args (which there are).
        # This has been fixed in Catkin Tools... perhaps we should be using
        # that instead.
        self._run_in_bash(catkincmd)

    def snap_fileset(self):
        """Filter useless files out of the snap.

        - opt/ros/<rosdistro>/.rosinstall points to the part installdir, and
          isn't useful from the snap anyway.
        """

        fileset = super().snap_fileset()
        fileset.append(
            "-{}".format(os.path.join("opt", "ros", self._rosdistro, ".rosinstall"))
        )
        return fileset


def _find_system_dependencies(catkin_packages, rosdep, rospack, catkin):
    """Find system dependencies for a given set of Catkin packages."""

    resolved_dependencies = {}
    dependencies = set()

    logger.info("Determining system dependencies for Catkin packages...")
    if catkin_packages is not None:
        for package in catkin_packages:
            # Query rosdep for the list of dependencies for this package
            dependencies |= rosdep.get_dependencies(package)
    else:
        # Rather than getting dependencies for an explicit list of packages,
        # let's get the dependencies for the entire workspace.
        dependencies |= rosdep.get_dependencies()

    if catkin_packages is None:
        catkin_packages = rospack.list_names()

    for dependency in dependencies:
        _resolve_package_dependencies(
            catkin_packages, dependency, catkin, rosdep, resolved_dependencies
        )

    # We currently have nested dict structure of:
    #    dependency name -> package type -> package names
    #
    # We want to return a flattened dict of package type -> package names.
    flattened_dependencies = {}
    for dependency_types in resolved_dependencies.values():
        for key, value in dependency_types.items():
            if key not in flattened_dependencies:
                flattened_dependencies[key] = set()
            flattened_dependencies[key] |= value

    # Finally, return that dict of dependencies
    return flattened_dependencies


def _resolve_package_dependencies(
    catkin_packages, dependency, catkin, rosdep, resolved_dependencies
):
    # No need to resolve this dependency if we know it's local, or if
    # we've already resolved it into a system dependency
    if dependency in resolved_dependencies or (
        catkin_packages and dependency in catkin_packages
    ):
        return

    if _dependency_is_in_underlay(catkin, dependency):
        # Package was found-- don't pull anything extra to satisfy
        # this dependency.
        logger.debug("Satisfied dependency {!r} in underlay".format(dependency))
        return

    # In this situation, the package depends on something that we
    # weren't instructed to build. It's probably a system dependency,
    # but the developer could have also forgotten to tell us to build
    # it.
    try:
        these_dependencies = rosdep.resolve_dependency(dependency)
    except _ros.rosdep.RosdepDependencyNotResolvedError:
        raise CatkinInvalidSystemDependencyError(dependency)

    for key, value in these_dependencies.items():
        if key not in _SUPPORTED_DEPENDENCY_TYPES:
            raise CatkinUnsupportedDependencyTypeError(key, dependency)

        resolved_dependencies[dependency] = {key: value}


def _dependency_is_in_underlay(catkin, dependency):
    if catkin:
        # Before trying to resolve this dependency into a system
        # dependency, see if it's already in the underlay.
        try:
            catkin.find(dependency)
        except CatkinPackageNotFoundError:
            # No package by that name is available
            pass
        else:
            return True
    return False


def _handle_rosinstall_files(wstool, rosinstall_files):
    """Merge given rosinstall files into our workspace."""

    for rosinstall_file in rosinstall_files:
        logger.info("Merging {}".format(rosinstall_file))
        wstool.merge(rosinstall_file)

    logger.info("Updating workspace...")
    wstool.update()


def _recursively_handle_rosinstall_files(wstool, source_path, *, cache=None):
    "Recursively find and merge rosinstall files and update workspace"

    rosinstall_files = set()  # type: Set[str]
    if not cache:
        cache = set()  # type: Set[str]

    # Walk the entire source directory looking for rosinstall files. Keep track
    # of any we haven't seen previously.
    for root, directories, files in os.walk(source_path):
        for file_name in files:
            path = os.path.join(root, file_name)
            if path.endswith(".rosinstall") and path not in cache:
                rosinstall_files.add(path)

    # If we came across previously-unseen rosinstall files, add them to the
    # cache. Then handle them (merge/update). Finally, walk again. Do this
    # until no new rosinstall files are discovered.
    if rosinstall_files:
        cache.update(rosinstall_files)
        _handle_rosinstall_files(wstool, rosinstall_files)
        _recursively_handle_rosinstall_files(wstool, source_path, cache=cache)


class CatkinPackageNotFoundError(errors.SnapcraftError):
    fmt = "Unable to find Catkin package {package_name!r}"

    def __init__(self, package_name):
        super().__init__(package_name=package_name)


class _Catkin:
    def __init__(
        self,
        ros_distro: str,
        workspaces: List[str],
        catkin_path: str,
        ubuntu_sources: str,
        ubuntu_keyrings: str,
        project: snapcraft.project.Project,
    ) -> None:
        self._ros_distro = ros_distro
        self._workspaces = workspaces
        self._catkin_path = catkin_path
        self._ubuntu_sources = ubuntu_sources
        self._ubuntu_keyrings = ubuntu_keyrings
        self._project = project
        self._catkin_install_path = os.path.join(self._catkin_path, "install")

    def setup(self):
        os.makedirs(self._catkin_install_path, exist_ok=True)

        # With the introduction of an underlay, we no longer know where Catkin
        # is. Let's just fetch/unpack our own, and use it.
        logger.info("Preparing to fetch catkin...")
        ubuntu = repo.Ubuntu(
            self._catkin_path,
            sources=self._ubuntu_sources,
            keyrings=self._ubuntu_keyrings,
            project_options=self._project,
        )
        logger.info("Fetching catkin...")
        ubuntu.get(["ros-{}-catkin".format(self._ros_distro)])

        logger.info("Installing catkin...")
        ubuntu.unpack(self._catkin_install_path)

    def find(self, package_name):
        with contextlib.suppress(subprocess.CalledProcessError):
            path = self._run(["--first-only", package_name]).strip()

            # Not a valid find if the package resolves into our own catkin
            # workspace. That won't be transitioned into the snap.
            if not path.startswith(self._catkin_install_path):
                return path

        raise CatkinPackageNotFoundError(package_name)

    def _run(self, arguments):
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            lines = [
                "export PYTHONPATH={}".format(
                    os.path.join(
                        self._catkin_install_path,
                        "usr",
                        "lib",
                        "python2.7",
                        "dist-packages",
                    )
                )
            ]

            ros_path = os.path.join(
                self._catkin_install_path, "opt", "ros", self._ros_distro
            )
            bin_paths = (
                os.path.join(ros_path, "bin"),
                os.path.join(self._catkin_install_path, "usr", "bin"),
            )
            lines.append(
                "export {}".format(
                    formatting_utils.format_path_variable(
                        "PATH", bin_paths, prepend="", separator=":"
                    )
                )
            )

            # Source our own workspace so we have all of Catkin's dependencies,
            # then source the workspace we're actually supposed to be crawling.
            lines.append(
                "_CATKIN_SETUP_DIR={} source {}".format(
                    ros_path, os.path.join(ros_path, "setup.sh")
                )
            )

            for workspace in self._workspaces:
                lines.append(
                    "_CATKIN_SETUP_DIR={} source {} --extend".format(
                        workspace, os.path.join(workspace, "setup.sh")
                    )
                )

            lines.append('exec "$@"')
            f.write("\n".join(lines))
            f.flush()
            return (
                subprocess.check_output(
                    ["/bin/bash", f.name, "catkin_find"] + arguments,
                    stderr=subprocess.STDOUT,
                )
                .decode("utf8")
                .strip()
            )


def _get_highest_version_path(path):
    paths = sorted(glob.glob(os.path.join(path, "*")))
    if not paths:
        raise CatkinNoHighestVersionPathError(path)

    return paths[-1]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import os
import functools
import operator
import stat
import sys
from textwrap import dedent
from typing import Dict, List, Union  # noqa: F401

import click
from tabulate import tabulate

import snapcraft
from snapcraft import storeapi, formatting_utils
from snapcraft.storeapi.constants import DEFAULT_SERIES
from . import echo


_MESSAGE_REGISTER_PRIVATE = dedent(
    """\
    Even though this is private snap, you should think carefully about
    the choice of name and make sure you are confident nobody else will
    have a stronger claim to that particular name. If you are unsure
    then we suggest you prefix the name with your developer identity,
    As ‘nessita-yoyodyne-www-site-content’."""
)
_MESSAGE_REGISTER_CONFIRM = dedent(
    """
    We always want to ensure that users get the software they expect
    for a particular name.

    If needed, we will rename snaps to ensure that a particular name
    reflects the software most widely expected by our community.

    For example, most people would expect ‘thunderbird’ to be published by
    Mozilla. They would also expect to be able to get other snaps of
    Thunderbird as 'thunderbird-$username'.

    Would you say that MOST users will expect {!r} to come from
    you, and be the software you intend to publish there?"""
)
_MESSAGE_REGISTER_SUCCESS = "Congrats! You are now the publisher of {!r}."
_MESSAGE_REGISTER_NO = dedent(
    """
    Thank you! {!r} will remain available.

    In the meantime you can register an alternative name."""
)


@contextlib.contextmanager
def _requires_login():
    try:
        yield
    except storeapi.errors.InvalidCredentialsError:
        echo.error("No valid credentials found." ' Have you run "snapcraft login"?')
        raise


@click.group()
def storecli():
    """Store commands"""
    pass


def _human_readable_acls(store: storeapi.StoreClient) -> str:
    acl = store.acl()
    snap_names = []
    if acl["snap_ids"]:
        for snap_id in acl["snap_ids"]:
            snap_names.append(store.get_snap_name_for_id(snap_id))
    acl["snap_names"] = snap_names

    human_readable_acl = {
        "expires": str(acl["expires"])
    }  # type: Dict[str, Union[str, List[str]]]

    for key in ("snap_names", "channels", "permissions"):
        human_readable_acl[key] = acl[key]
        if not acl[key]:
            human_readable_acl[key] = "No restriction"

    return dedent(
        """\
        snaps:       {snap_names}
        channels:    {channels}
        permissions: {permissions}
        expires:     {expires}
    """.format(
            **human_readable_acl
        )
    )


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.option("--private", is_flag=True, help="Register the snap as a private one")
@click.option("--store", metavar="<store>", help="Store to register with")
def register(snap_name, private, store):
    """Register <snap-name> with the store.

    You can use this command to register an available <snap-name> and become
    the publisher for this snap.

    \b
    Examples:
        snapcraft register thunderbird
    """
    if private:
        click.echo(_MESSAGE_REGISTER_PRIVATE.format(snap_name))
    if echo.confirm(_MESSAGE_REGISTER_CONFIRM.format(snap_name)):
        snapcraft.register(snap_name, is_private=private, store_id=store)
        click.echo(_MESSAGE_REGISTER_SUCCESS.format(snap_name))
    else:
        click.echo(_MESSAGE_REGISTER_NO.format(snap_name))


@storecli.command()
@click.option(
    "--release",
    metavar="<channels>",
    help="Optional comma separated list of channels to release <snap-file>",
)
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
def push(snap_file, release):
    """Push <snap-file> to the store.

    By passing --release with a comma separated list of channels the snap would
    be released to the selected channels if the store review passes for this
    <snap-file>.

    This operation will block until the store finishes processing this
    <snap-file>.

    If --release is used, the channel map will be displayed after the
    operation takes place.

    \b
    Examples:
        snapcraft push my-snap_0.1_amd64.snap
        snapcraft push my-snap_0.2_amd64.snap --release edge
        snapcraft push my-snap_0.3_amd64.snap --release candidate,beta
    """
    click.echo("Preparing to push {!r}.".format(os.path.basename(snap_file)))
    channel_list = []
    if release:
        channel_list = release.split(",")
        click.echo(
            "After pushing, an attempt will be made to release to {}"
            "".format(formatting_utils.humanize_list(channel_list, "and"))
        )

    snapcraft.push(snap_file, channel_list)


@storecli.command("push-metadata")
@click.option(
    "--force",
    is_flag=True,
    help="Force metadata update to override any possible conflict",
)
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
def push_metadata(snap_file, force):
    """Push metadata from <snap-file> to the store.

    The following information will be retrieved from <snap-file> and used
    to update the store:

    \b
    - summary
    - description
    - icon

    If --force is given, it will force the local metadata into the Store,
    ignoring any possible conflict.

    \b
    Examples:
        snapcraft push-metadata my-snap_0.1_amd64.snap
        snapcraft push-metadata my-snap_0.1_amd64.snap --force
    """
    click.echo("Pushing metadata from {!r}".format(os.path.basename(snap_file)))
    snapcraft.push_metadata(snap_file, force)


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("revision", metavar="<revision>")
@click.argument("channels", metavar="<channels>")
def release(snap_name, revision, channels):
    """Release <snap-name> on <revision> to the selected store <channels>.
    <channels> is a comma separated list of valid channels on the
    store.

    The <revision> must exist on the store, to see available revisions
    run `snapcraft list-revisions <snap_name>`.

    The channel map will be displayed after the operation takes place.
    To see the status map at any other time run `snapcraft status <snap-name>`.

    The format for channels is `[<track>/]<risk>[/<branch>]` where

    \b
        - <track> is used to have long term release channels. It is implicitly
          set to `latest`. If this snap requires one, it can be created by
          request by having a conversation on https://forum.snapcraft.io
          under the *store* category.
        - <risk> is mandatory and can be either `stable`, `candidate`, `beta`
          or `edge`.
        - <branch> is optional and dynamically creates a channel with a
          specific expiration date.

    \b
    Examples:
        snapcraft release my-snap 8 stable
        snapcraft release my-snap 8 stable/my-branch
        snapcraft release my-snap 9 beta,edge
        snapcraft release my-snap 9 lts-channel/stable
        snapcraft release my-snap 9 lts-channel/stable/my-branch
    """
    snapcraft.release(snap_name, revision, channels.split(","))


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.option(
    "--from-channel",
    metavar="<from-channel>",
    required=True,
    help="The channel to promote from.",
)
@click.option(
    "--to-channel",
    metavar="<to-channel>",
    required=True,
    help="The channel to promote to.",
)
@click.option("--yes", is_flag=True, help="Do not prompt for confirmation.")
def promote(snap_name, from_channel, to_channel, yes):
    """Promote a build set from to a channel.

    A build set is a set of commonly tagged revisions, the most simple
    form of a build set is a set of revisions released to a channel.

    Currently, only channels are supported to release from (<from-channel>)

    Prior to releasing, visual confirmation shall be required.

    The format for channels is `[<track>/]<risk>[/<branch>]` where

    \b
        - <track> is used to have long term release channels. It is implicitly
          set to the default.
        - <risk> is mandatory and can be either `stable`, `candidate`, `beta`
          or `edge`.
        - <branch> is optional and dynamically creates a channel with a
          specific expiration date.

    \b
    Examples:
        snapcraft promote my-snap --from-channel candidate --to-channel stable
        snapcraft promote my-snap --from-channel lts/candidate --to-channel lts/stable
        snapcraft promote my-snap --from-channel stable/patch --to-channel stable
        snapcraft promote my-snap --from-channel experimental/stable --to-channel stable
    """
    echo.warning(
        "snapcraft promote does not have a stable CLI interface. Use with caution in scripts."
    )
    parsed_from_channel = storeapi.channels.Channel(from_channel)
    parsed_to_channel = storeapi.channels.Channel(to_channel)

    if parsed_from_channel == parsed_to_channel:
        raise click.BadOptionUsage(
            "--to-channel", "--from-channel and --to-channel cannot be the same."
        )
    elif parsed_from_channel.risk == "edge" and parsed_from_channel.branch is None:
        raise click.BadOptionUsage(
            "--from-channel",
            "{!r} is not a valid set value for --from-channel.".format(
                parsed_from_channel
            ),
        )

    store = storeapi.StoreClient()
    with _requires_login():
        status_payload = store.get_snap_status(snap_name)

    snap_status = storeapi.status.SnapStatus(
        snap_name=snap_name, payload=status_payload
    )
    from_channel_set = snap_status.get_channel_set(parsed_from_channel)
    echo.info("Build set information for {!r}".format(parsed_from_channel))
    click.echo(
        tabulate(
            sorted(from_channel_set, key=operator.attrgetter("arch")),
            headers=["Arch", "Revision", "Version"],
            tablefmt="plain",
        )
    )
    if yes or echo.confirm(
        "Do you want to promote the current set to the {!r} channel?".format(
            parsed_to_channel
        )
    ):
        for c in from_channel_set:
            snapcraft.release(snap_name, str(c.revision), [str(parsed_to_channel)])
    else:
        echo.wrapped("Channel promotion cancelled")


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("channels", metavar="<channel>...", nargs=-1)
def close(snap_name, channels):
    """Close <channel> for <snap-name>.
    Closing a channel allows the <channel> that is closed to track the channel
    that follows it in the channel release chain. As such closing the
    'candidate' channel would make it track the 'stable' channel.

    The channel map will be displayed after the operation takes place.

    \b
    Examples:
        snapcraft close my-snap beta
        snapcraft close my-snap beta edge
    """
    snapcraft.close(snap_name, channels)


@storecli.command()
@click.option(
    "--arch", metavar="<arch>", help="The snap architecture to get the status for"
)
@click.option(
    "--series",
    metavar="<series>",
    default=DEFAULT_SERIES,
    help="The snap series to get the status for",
)
@click.argument("snap-name", metavar="<snap-name>")
def status(snap_name, series, arch):
    """Get the status on the store for <snap-name>.

    \b
    Examples:
        snapcraft status my-snap
        snapcraft status my-snap --arch armhf
    """
    snapcraft.status(snap_name, series, arch)


@storecli.command("list-revisions")
@click.option(
    "--arch", metavar="<arch>", help="The snap architecture to get the status for"
)
@click.option(
    "--series",
    metavar="<series>",
    default=DEFAULT_SERIES,
    help="The snap series to get the status for",
)
@click.argument("snap-name", metavar="<snap-name>")
def list_revisions(snap_name, series, arch):
    """Get the history on the store for <snap-name>.

    This command has an alias of `revisions`.

    \b
    Examples:
        snapcraft list-revisions my-snap
        snapcraft list-revisions my-snap --arch armhf
        snapcraft revisions my-snap
    """
    snapcraft.revisions(snap_name, series, arch)


@storecli.command("list-registered")
def list_registered():
    """List snap names registered or shared with you.

    This command has an alias of `registered`.

    \b
    Examples:
        snapcraft list-registered
        snapcraft registered
    """
    snapcraft.list_registered()


@storecli.command("export-login")
@click.argument(
    "login_file", metavar="FILE", type=click.Path(dir_okay=False, writable=True)
)
@click.option(
    "--snaps", metavar="<snaps>", help="Comma-separated list of snaps to limit access"
)
@click.option(
    "--channels",
    metavar="<channels>",
    help="Comma-separated list of channels to limit access",
)
@click.option(
    "--acls", metavar="<acls>", help="Comma-separated list of ACLs to limit access"
)
@click.option(
    "--expires",
    metavar="<expiration date>",
    help="Date/time (in ISO 8601) when this exported login expires",
)
def export_login(login_file: str, snaps: str, channels: str, acls: str, expires: str):
    """Save login configuration for a store account in FILE.

    This file can then be used to log in to the given account with the
    specified permissions. One can also request the login to be exported to
    stdout instead of a file:

        snapcraft export-login -

    For example, to limit access to the edge channel of any snap the account
    can access:

        snapcraft export-login --channels=edge exported

    Or to limit access to only the edge channel of a single snap:

        snapcraft export-login --snaps=my-snap --channels=edge exported

    To limit access to a single snap, but only until 2019:

        snapcraft export-login --expires="2019-01-01T00:00:00" exported
    """

    snap_list = None
    channel_list = None
    acl_list = None

    if snaps:
        snap_list = []
        for package in snaps.split(","):
            snap_list.append({"name": package, "series": "16"})

    if channels:
        channel_list = channels.split(",")

    if acls:
        acl_list = acls.split(",")

    store = storeapi.StoreClient()
    if not snapcraft.login(
        store=store,
        packages=snap_list,
        channels=channel_list,
        acls=acl_list,
        expires=expires,
        save=False,
    ):
        sys.exit(1)

    # Support a login_file of '-', which indicates a desire to print to stdout
    if login_file.strip() == "-":
        echo.info("\nExported login starts on next line:")
        store.conf.save(config_fd=sys.stdout, encode=True)
        print()

        preamble = "Login successfully exported and printed above"
        login_action = 'echo "<login>" | snapcraft login --with -'
    else:
        # This is sensitive-- it should only be accessible by the owner
        private_open = functools.partial(os.open, mode=0o600)

        # mypy doesn't have the opener arg in its stub. Ignore its warning
        with open(login_file, "w", opener=private_open) as f:  # type: ignore
            store.conf.save(config_fd=f)

        # Now that the file has been written, we can just make it
        # owner-readable
        os.chmod(login_file, stat.S_IRUSR)

        preamble = "Login successfully exported to {0!r}".format(login_file)
        login_action = "snapcraft login --with {0}".format(login_file)

    print()
    echo.info(
        dedent(
            """\
        {}. This can now be used with

            {}

        to log in to this account with no password and have these
        capabilities:\n""".format(
                preamble, login_action
            )
        )
    )
    echo.info(_human_readable_acls(store))
    echo.warning(
        "This exported login is not encrypted. Do not commit it to version control!"
    )


@storecli.command()
@click.option(
    "--with",
    "login_file",
    metavar="<login file>",
    type=click.File("r"),
    help="Path to file created with 'snapcraft export-login'",
)
def login(login_file):
    """Login with your Ubuntu One e-mail address and password.

    If you do not have an Ubuntu One account, you can create one at
    https://snapcraft.io/account
    """
    store = storeapi.StoreClient()
    if not snapcraft.login(store=store, config_fd=login_file):
        sys.exit(1)

    print()

    if login_file:
        echo.info("Login successful. You now have these capabilities:\n")
        echo.info(_human_readable_acls(store))
    else:
        echo.info("Login successful.")


@storecli.command()
def logout():
    """Clear session credentials."""
    store = storeapi.StoreClient()
    store.logout()
    echo.info("Credentials cleared.")


@storecli.command()
def whoami():
    """Returns your login information relevant to the store."""
    try:
        account_data = storeapi.StoreClient().whoami()
    except storeapi.errors.InvalidCredentialsError:
        echo.error("You need to first login to use this command.")
        sys.exit(1)

    click.echo(
        dedent(
            """\
        email:        {email}
        developer-id: {account_id}""".format(
                **account_data
            )
        )
    )

    # This is needed because we originally did not store the login information.
    if account_data["email"] == "unknown":
        echo.warning(
            "In order to view the correct email you will need to "
            "logout and login again."
        )
